
#include <stdio.h>
#include <stdlib.h>
#include "ch3d.h"

#ifndef	NO_SFILT
#define FLOAT	double
#define EPSILON	1.0e-13
#endif

static void	Error(char *msg)
{
	(void)fputs(msg,stderr); (void)fputc('\n',stderr); exit(1);
}

#define X	CH3D_X(ch3d)
#define Y	CH3D_Y(ch3d)
#define Z	CH3D_Z(ch3d)
#define PARA1	(ch3d->para1)
#define PARA2	(ch3d->para2)
#define PARA3	(ch3d->para3)
#define PARA4	(ch3d->para4)
#define PARA5	(ch3d->para5)
#define NAME	(ch3d->name)
#define VE	(ch3d->ve)
#define FE	(ch3d->fe)
#define SV	(ch3d->sv)
#define EV	(ch3d->ev)
#define RF	(ch3d->rf)
#define LF	(ch3d->lf)
#define SCE	(ch3d->sce)
#define SCCE	(ch3d->scce)
#define ECE	(ch3d->ece)
#define ECCE	(ch3d->ecce)
#define WV1	(ch3d->wv1)
#define WV2	(ch3d->wv2)
#define WV3	(ch3d->wv3)
#define WV4	(ch3d->wv4)
#define WV5	(ch3d->wv5)
#define WV6	(ch3d->wv6)
#define WE1	(ch3d->we1)
#define WE2	(ch3d->we2)
#define WF1	(ch3d->wf1)
#define WF2	(ch3d->wf2)

#define CLIST	WV1
#define MOL	WV2
#define VSTAC	WV2
#define HULNV	WV3
#define HULPV	WV4
#define CNXV	WV5
#define CCNXV	WV6
#define ECOLO	WE1
#define ESTAC	WE2
#define FCOLO	WF1
#define FQUE	WF2

CH3D_T	*InitCH3D(CH3D_U Np)
{
	CH3D_U			n3=Np*3,n4=Np*4;
	unsigned long long	n=(unsigned long long)Np+
				  Np+n3+n4+n4+n4+n4+n4+n4+n4+n4+
				  Np+Np+Np+Np+Np+Np+n4+n4+n3+n3;
	CH3D_T			*ch3d;

	if (Np<4 || Np>CH3D_LIM/4 || n>(~(size_t)0)/sizeof(size_t) ||
	    (ch3d=(CH3D_T *)malloc(sizeof(CH3D_T)))==NULL ||
	    (X=(CH3D_I *)malloc(sizeof(CH3D_I)*Np))==NULL ||
	    (Y=(CH3D_I *)malloc(sizeof(CH3D_I)*Np))==NULL ||
	    (Z=(CH3D_I *)malloc(sizeof(CH3D_I)*Np))==NULL ||
	    (NAME=(CH3D_U *)malloc(sizeof(CH3D_U)*(size_t)n))==NULL)
	    Error("no applicable memory for CH3D.");

	WF2=(WF1=(WE2=(WE1=(WV6=(WV5=(WV4=(WV3=(WV2=(WV1=(ECCE=(
	ECE=(SCCE=(SCE=(LF=(RF=(EV=(SV=(FE=(VE=NAME+Np)+
	Np)+n3)+n4)+n4)+n4)+n4)+n4)+n4)+n4)+n4)+
	Np)+Np)+Np)+Np)+Np)+Np)+n4)+n4)+n3;
	PARA1=Np; PARA2=n3; PARA3=n4;
	return ch3d;
}

void	TermCH3D(CH3D_T *ch3d)
{
	free(NAME); free(Z); free(Y); free(X); free(ch3d);
}

static void	Sort_NAME(CH3D_T *ch3d,CH3D_U head,CH3D_U tail)
{
	CH3D_U		u,*mol=MOL-1,n=NAME[mol[head+(tail-head)/2]];
	long long	h=head,t=tail;

	do {
	    while (NAME[mol[h]]<n) ++h;
	    while (n<NAME[mol[t]]) --t;

	    if (h<=t) {
		u=mol[h]; mol[h]=mol[t]; mol[t]=u; ++h; --t;
	    }
	} while (h<=t);

	if (head<t) Sort_NAME(ch3d,head,(CH3D_U)t);
	if (h<tail) Sort_NAME(ch3d,(CH3D_U)h,tail);
}

#define LARGER(v1,v2,x,name)	\
	(x[v1]>x[v2] || (x[v1]==x[v2] && name[v1]<name[v2]))

static void	Sort_X(CH3D_T *ch3d,CH3D_U head,CH3D_U tail)
{
	CH3D_U		u,*mol=MOL-1,m=mol[head+(tail-head)/2];
	long long	h=head,t=tail;

	do {
	    while (LARGER(m,mol[h],X,NAME)) ++h;
	    while (LARGER(mol[t],m,X,NAME)) --t;

	    if (h<=t) {
		u=mol[h]; mol[h]=mol[t]; mol[t]=u; ++h; --t;
	    }
	} while (h<=t);

	if (head<t) Sort_X(ch3d,head,(CH3D_U)t);
	if (h<tail) Sort_X(ch3d,(CH3D_U)h,tail);
}

static int	SORTBN(int nofp,CH3D_U *list1,CH3D_U *name,CH3D_U *list2)
{
	int	j,k,i=1;
	CH3D_U	l;

	if (list2==NULL)
	    list2=list1;
	else
	    for (j=0; j<nofp; j++) list2[j]=list1[j];

	for (j=nofp-1; j>0; j--)
	    for (k=0; k<j; k++)
		if (name[list2[k]]>name[list2[k+1]]) {
		    l=list2[k]; list2[k]=list2[k+1]; list2[k+1]=l; i=(-i);
		}

	return i;
}

#define BYTE	unsigned char
#define WORD	unsigned int
#define SIZE	sizeof(CH3D_I)

#define U_I	unsigned CH3D_I
#define MSB	(sizeof(WORD)*8-1)

static void	Mul_2(CH3D_I a,CH3D_I b,BYTE *C)
{
	WORD	x,y,z;
	U_I	u,v,w;

	for (x=0; x<SIZE; x++) C[x]=0;

	for (u=(U_I)b, y=0; y<SIZE; y++, u>>=8) {
	    v=u&255;
	    for (w=(U_I)a, z=x=0; x<SIZE; x++, z>>=8, w>>=8)
		C[x+y]=(BYTE)(z+=((w&255)*v+C[x+y]));

	    C[SIZE+y]=(BYTE)z;
	}
	if (a<0)
	    for (u=(U_I)b, z=y=0; y<SIZE; y++, z>>=MSB, u>>=8)
		C[SIZE+y]=(BYTE)(z=C[SIZE+y]-(u&255)-z);

	if (b<0)
	    for (w=(U_I)a, z=x=0; x<SIZE; x++, z>>=MSB, w>>=8)
		C[x+SIZE]=(BYTE)(z=C[x+SIZE]-(w&255)-z);
}

static void	Mul_3(CH3D_I a,CH3D_I b,CH3D_I c,BYTE *D)
{
	BYTE	C[SIZE*2];
	WORD	x,y,z;
	U_I	w;

	Mul_2(a,b,C);

	for (x=0; x<SIZE; x++) D[x]=0;

	for (y=0; y<SIZE*2; y++) {
	    for (w=(U_I)c, z=x=0; x<SIZE; x++, z>>=8, w>>=8)
		D[x+y]=(BYTE)(z+=((w&255)*C[y]+D[x+y]));

	    D[SIZE+y]=(BYTE)z;
	}
	if (c<0)
	    for (z=y=0; y<SIZE*2; y++, z>>=MSB)
		D[SIZE+y]=(BYTE)(z=D[SIZE+y]-C[y]-z);

	if ((char)C[SIZE*2-1]<0)
	    for (w=(U_I)c, z=x=0; x<SIZE; x++, z>>=MSB, w>>=8)
		D[x+SIZE*2]=(BYTE)(z=D[x+SIZE*2]-(w&255)-z);
}

#define SIGN(s,E)	E[s]=((char)E[s-1]<0)?255:0

static void	Add(size_t s,BYTE *A,BYTE *B)
{
	WORD	y,z=0;

	for (y=0; y<=s; y++, z>>=8) B[y]=(BYTE)(z+=((WORD)A[y]+B[y]));
}

static void	Sub(size_t s,BYTE *A,BYTE *B)
{
	WORD	y,z=1;

	for (y=0; y<=s; y++, z>>=8) B[y]=(BYTE)(z+=((WORD)A[y]+(BYTE)(~B[y])));
}

static int	Sign(size_t s,BYTE *C)
{
	if ((char)C[s]<0) return -1;

	while (s!=0 && C[s]==0) --s; return C[s]!=0;
}

static void	Det(CH3D_U v0,CH3D_U v1,CH3D_U v2,
		    CH3D_I *x,CH3D_I *y,CH3D_I *z,BYTE *D)
{
	BYTE	C[SIZE*3+1];

#define MUL_3(a,b,c,ABC)	Mul_3(a,b,c,ABC); SIGN(SIZE*3,ABC)

	MUL_3(x[v0],y[v1],z[v2],C);
	MUL_3(x[v1],y[v2],z[v0],D); Add(SIZE*3,C,D);
	MUL_3(x[v2],y[v0],z[v1],C); Add(SIZE*3,D,C);
	MUL_3(x[v0],y[v2],z[v1],D); Sub(SIZE*3,C,D);
	MUL_3(x[v1],y[v0],z[v2],C); Sub(SIZE*3,D,C);
	MUL_3(x[v2],y[v1],z[v0],D); Sub(SIZE*3,C,D);
}

static int	JD1SSS(CH3D_U *v,CH3D_I *x,CH3D_I *y,CH3D_I *z)
{
	BYTE	A[SIZE*3+1],B[SIZE*3+1];

	Det(v[1],v[2],v[3],x,y,z,A);
	Det(v[0],v[2],v[3],x,y,z,B); Sub(SIZE*3,A,B);
	Det(v[0],v[1],v[3],x,y,z,A); Add(SIZE*3,B,A);
	Det(v[0],v[1],v[2],x,y,z,B); Sub(SIZE*3,A,B); return Sign(SIZE*3,B);
}

static int	JSD1SS(CH3D_U v0,CH3D_U v1,CH3D_U v2,CH3D_I *x,CH3D_I *y)
{
	BYTE	A[SIZE*2+1],B[SIZE*2+1];

#define MUL_2(a,b,AB)	Mul_2(a,b,AB); SIGN(SIZE*2,AB)

	MUL_2(x[v0],y[v1],A);
	MUL_2(x[v1],y[v2],B); Add(SIZE*2,A,B);
	MUL_2(x[v2],y[v0],A); Add(SIZE*2,B,A);
	MUL_2(x[v0],y[v2],B); Sub(SIZE*2,A,B);
	MUL_2(x[v1],y[v0],A); Sub(SIZE*2,B,A);
	MUL_2(x[v2],y[v1],B); Sub(SIZE*2,A,B); return Sign(SIZE*2,B);
}

#define JSD1S(v1,v2,x)	((x[v1]>x[v2])?-1:x[v1]<x[v2])

static int	JORSSS(CH3D_U *v,CH3D_I *x,CH3D_I *y,CH3D_I *z,CH3D_U *name)
{
	CH3D_U	list[4];
	int	sign,change=SORTBN(4,v,name,list);

#ifndef	NO_SFILT
	CH3D_U	v0=list[0],v1=list[1],v2=list[2],v3=list[3];
	FLOAT	d,f,
		e=1.0,
		x0=(FLOAT)x[v0],y0=(FLOAT)y[v0],z0=(FLOAT)z[v0],
		x1=(FLOAT)x[v1]-x0,y1=(FLOAT)y[v1]-y0,z1=(FLOAT)z[v1]-z0,
		x2=(FLOAT)x[v2]-x0,y2=(FLOAT)y[v2]-y0,z2=(FLOAT)z[v2]-z0,
		x3=(FLOAT)x[v3]-x0,y3=(FLOAT)y[v3]-y0,z3=(FLOAT)z[v3]-z0;

	d = f=x1*y2*z3 ; if (f<0.0) f=(-f); if (f>e) e=f;
	d+=(f=x2*y3*z1); if (f<0.0) f=(-f); if (f>e) e=f;
	d+=(f=x3*y1*z2); if (f<0.0) f=(-f); if (f>e) e=f;
	d-=(f=x3*y2*z1); if (f<0.0) f=(-f); if (f>e) e=f;
	d-=(f=x2*y1*z3); if (f<0.0) f=(-f); if (f>e) e=f;
	d-=(f=x1*y3*z2); if (f<0.0) f=(-f); if (f>e) e=f; e*=EPSILON;

	if (!(sign=(d<-e)?-1:d>e))
#endif

/*  0 */    if (!(sign=  JD1SSS(list,x,y,z)))
/*  1 */    if (!(sign= -JSD1SS(list[1],list[2],list[3],y,z)))
/*  2 */    if (!(sign=  JSD1SS(list[0],list[2],list[3],y,z)))
/*  3 */    if (!(sign= -JSD1SS(list[0],list[1],list[3],y,z)))
/*  4 */    if (!(sign=  JSD1SS(list[1],list[2],list[3],x,z)))
/*  5 */    if (!(sign= -JSD1S (list[2],list[3],z)))
/*  6 */    if (!(sign=  JSD1S (list[1],list[3],z)))
/*  7 */    if (!(sign= -JSD1SS(list[0],list[2],list[3],x,z)))
/*  8 */    if (!(sign= -JSD1S (list[0],list[3],z)))
/*  9 */    if (!(sign=  JSD1SS(list[0],list[1],list[3],x,z)))
/* 10 */    if (!(sign= -JSD1SS(list[1],list[2],list[3],x,y)))
/* 11 */    if (!(sign=  JSD1S (list[2],list[3],y)))
/* 12 */    if (!(sign= -JSD1S (list[1],list[3],y)))
/* 13 */    if (!(sign= -JSD1S (list[2],list[3],x))) sign=1;

	return sign*change;
}

static CH3D_U	KNEWF(CH3D_T *ch3d)
{
	CH3D_U	u=PARA4;

	if (u==CH3D_LIM) Error("too many faces.");

	PARA4=FE[u]; return u;
}

static CH3D_U	KNEWE(CH3D_T *ch3d)
{
	CH3D_U	u=PARA5;

	if (u==CH3D_LIM) Error("too many edges.");

	PARA5=SV[u]; return u;
}

typedef struct {
		CH3D_U	sv,ev,rf,lf,sce,scce,ece,ecce;
	} WES;

static void     Set(CH3D_T *ch3d,CH3D_U Nv,CH3D_U *v,
				 CH3D_U Nf,CH3D_U *f,
				 CH3D_U Ne,CH3D_U *e,WES *wes)
{
	CH3D_U	u,w;

	for (u=0; u<Nv; u++) {
	    HULNV[v[u]]=v[(u+1)%Nv]; HULPV[v[u]]=v[(u+Nv-1)%Nv];
	}
	for (u=0; u<Nf; u++) f[u]=KNEWF(ch3d);
	for (u=0; u<Ne; u++) e[u]=KNEWE(ch3d);
	for (u=0; u<Ne; u++) {
	    w=e[u]; VE[SV[w]=v[wes[u].sv]]=VE[EV[w]=v[wes[u].ev]]=
		    FE[RF[w]=f[wes[u].rf]]=FE[LF[w]=f[wes[u].lf]]=w;
		    SCE[w]=e[wes[u].sce]; SCCE[w]=e[wes[u].scce];
		    ECE[w]=e[wes[u].ece]; ECCE[w]=e[wes[u].ecce];
	}
}

static WES	R_0[6]={{0,1,0,1,2,3,4,1},{1,2,0,2,0,4,5,2},{2,0,0,3,1,5,3,0},
			{0,3,1,3,0,2,5,4},{3,1,1,2,3,5,1,0},{3,2,2,3,4,3,2,1}},
		L_0[6]={{0,1,1,0,3,2,1,4},{1,2,2,0,4,0,2,5},{2,0,3,0,5,1,0,3},
			{0,3,3,1,2,0,4,5},{3,1,2,1,5,3,0,1},{3,2,3,2,3,4,1,2}},

		R_2[9]={{0,1,0,1,2,6,7,1},{1,2,0,2,0,4,5,2},{2,0,0,3,1,5,3,0},
			{0,3,4,3,6,2,5,8},{3,1,5,2,8,5,1,7},{3,2,2,3,4,3,2,1},
			{0,4,1,4,0,3,8,7},{4,1,1,5,6,8,4,0},{3,4,4,5,3,4,7,6}},
		R_4[9]={{0,1,0,1,2,3,4,1},{1,2,0,2,0,6,7,2},{2,0,0,3,1,5,3,0},
			{0,3,1,3,0,2,5,4},{3,1,1,4,3,8,6,0},{3,2,5,3,8,3,2,7},
			{1,4,2,4,1,4,8,7},{4,2,2,5,6,8,5,1},{3,4,4,5,4,5,7,6}},
		R_8[9]={{0,1,0,1,2,3,4,1},{1,2,0,2,0,4,5,2},{2,0,0,3,1,6,7,0},
			{0,3,1,5,0,7,8,4},{3,1,1,2,3,5,1,0},{3,2,2,4,4,8,6,1},
			{2,4,3,4,2,5,8,7},{4,0,3,5,6,8,3,2},{3,4,4,5,5,3,7,6}},
		R_5[9]={{0,1,0,1,6,3,4,1},{1,4,0,2,0,4,7,6},{2,0,5,3,8,5,3,6},
			{0,3,1,3,0,2,5,4},{3,1,1,2,3,7,1,0},{3,2,4,3,7,3,2,8},
			{4,0,0,5,1,8,2,0},{3,4,2,4,4,5,8,1},{2,4,4,5,5,2,6,7}},
		R_9[9]={{0,1,5,1,8,3,4,6},{1,2,0,2,6,4,5,2},{2,4,0,3,1,5,7,6},
			{0,3,1,4,0,8,7,4},{3,1,1,2,3,5,1,0},{3,2,2,3,4,7,2,1},
			{4,1,0,5,2,8,0,1},{3,4,3,4,5,3,8,2},{0,4,4,5,3,0,6,7}},
		R_6[9]={{0,1,0,5,2,6,8,1},{1,2,0,4,0,8,7,2},{2,0,0,3,1,5,3,0},
			{0,3,1,3,6,2,5,4},{3,4,1,2,3,5,7,6},{3,2,2,3,4,3,2,7},
			{4,0,1,5,4,8,0,3},{2,4,2,4,5,1,8,4},{1,4,4,5,1,0,6,7}},
		R_a[9]={{0,1,0,1,2,3,6,1},{1,2,0,2,0,4,5,2},{2,0,0,3,1,7,3,0},
			{0,4,1,3,0,2,7,6},{3,1,5,2,8,5,1,6},{3,2,2,4,4,8,7,1},
			{4,1,1,5,3,8,4,0},{2,4,3,4,2,5,8,3},{3,4,4,5,5,4,6,7}},
		R_c[9]={{0,1,0,1,2,3,4,1},{1,2,0,5,0,6,8,2},{2,0,0,4,1,8,7,0},
			{0,3,1,3,0,7,5,4},{3,1,1,2,3,5,6,0},{3,4,2,3,4,3,7,6},
			{4,1,2,5,5,8,1,4},{0,4,3,4,3,2,8,5},{2,4,4,5,2,1,6,7}},

		L_2[9]={{0,1,1,0,6,2,1,7},{1,2,2,0,4,0,2,5},{2,0,3,0,5,1,0,3},
			{0,3,3,4,2,6,8,5},{3,1,2,5,5,8,7,1},{3,2,3,2,3,4,1,2},
			{0,4,4,1,3,0,7,8},{4,1,5,1,8,6,0,4},{3,4,5,4,4,3,6,7}},
		L_4[9]={{0,1,1,0,3,2,1,4},{1,2,2,0,6,0,2,7},{2,0,3,0,5,1,0,3},
			{0,3,3,1,2,0,4,5},{3,1,4,1,8,3,0,6},{3,2,3,5,3,8,7,2},
			{1,4,4,2,4,1,7,8},{4,2,5,2,8,6,1,5},{3,4,5,4,5,4,6,7}},
		L_8[9]={{0,1,1,0,3,2,1,4},{1,2,2,0,4,0,2,5},{2,0,3,0,6,1,0,7},
			{0,3,5,1,7,0,4,8},{3,1,2,1,5,3,0,1},{3,2,4,2,8,4,1,6},
			{2,4,4,3,5,2,7,8},{4,0,5,3,8,6,2,3},{3,4,5,4,3,5,6,7}},
		L_5[9]={{0,1,1,0,3,6,1,4},{1,4,2,0,4,0,6,7},{2,0,3,5,5,8,6,3},
			{0,3,3,1,2,0,4,5},{3,1,2,1,7,3,0,1},{3,2,3,4,3,7,8,2},
			{4,0,5,0,8,1,0,2},{3,4,4,2,5,4,1,8},{2,4,5,4,2,5,7,6}},
		L_9[9]={{0,1,1,5,3,8,6,4},{1,2,2,0,4,6,2,5},{2,4,3,0,5,1,6,7},
			{0,3,4,1,8,0,4,7},{3,1,2,1,5,3,0,1},{3,2,3,2,7,4,1,2},
			{4,1,5,0,8,2,1,0},{3,4,4,3,3,5,2,8},{0,4,5,4,0,3,7,6}},
		L_6[9]={{0,1,5,0,6,2,1,8},{1,2,4,0,8,0,2,7},{2,0,3,0,5,1,0,3},
			{0,3,3,1,2,6,4,5},{3,4,2,1,5,3,6,7},{3,2,3,2,3,4,7,2},
			{4,0,5,1,8,4,3,0},{2,4,4,2,1,5,4,8},{1,4,5,4,0,1,7,6}},
		L_a[9]={{0,1,1,0,3,2,1,6},{1,2,2,0,4,0,2,5},{2,0,3,0,7,1,0,3},
			{0,4,3,1,2,0,6,7},{3,1,2,5,5,8,6,1},{3,2,4,2,8,4,1,7},
			{4,1,5,1,8,3,0,4},{2,4,4,3,5,2,3,8},{3,4,5,4,4,5,7,6}},
		L_c[9]={{0,1,1,0,3,2,1,4},{1,2,5,0,6,0,2,8},{2,0,4,0,8,1,0,7},
			{0,3,3,1,7,0,4,5},{3,1,2,1,5,3,0,6},{3,4,3,2,3,4,6,7},
			{4,1,5,2,8,5,4,1},{0,4,4,3,2,3,5,8},{2,4,5,4,1,2,7,6}};

#define TETRA_R()	Set(ch3d,4,v,4,f,6,e,R_0)
#define TETRA_L()	Set(ch3d,4,v,4,f,6,e,L_0)
#define TRI_DIP(R_L)	Set(ch3d,5,v,6,f,9,e,R_L)

static CH3D_U	CW_V_T[4][3]={{0,2,1},{1,3,0},{2,3,1},{0,3,2}};

static void	CH5P(CH3D_T *ch3d)
{
	CH3D_U	v[5],u,p,f[6],e[9];
	int	o[4];

	v[3]=MOL[4];
	for (u=0; u<4; u++) {
	    for (p=0; p<3; p++) v[p]=MOL[CW_V_T[u][p]];

	    o[u]=JORSSS(v,X,Y,Z,NAME);
	}
	for (u=0; u<5; u++) v[u]=MOL[u];

	if (JORSSS(v,X,Y,Z,NAME)>0)
	    switch((o[0]>0)|((o[1]>0)*2)|((o[2]>0)*4)|(o[3]>0)*8) {
		case 0x2: TRI_DIP(R_2); break;
		case 0x4: TRI_DIP(R_4); break;
		case 0x8: TRI_DIP(R_8); break;
		case 0x5: TRI_DIP(R_5); break;
		case 0x9: TRI_DIP(R_9); break;
		case 0x6: TRI_DIP(R_6); break;
		case 0xa: TRI_DIP(R_a); break;
		case 0xc: TRI_DIP(R_c); break;
		case 0x7: v[1]=v[4]; TETRA_R(); break;
		case 0xd: v[2]=v[4]; TETRA_R(); break;
		case 0xe: v[3]=v[4]; TETRA_R(); break;
		default : Error("unexpected right-handed system in CH5P.");
	    }
	else
	    switch((o[0]<0)|((o[1]<0)*2)|((o[2]<0)*4)|(o[3]<0)*8) {
		case 0x2: TRI_DIP(L_2); break;
		case 0x4: TRI_DIP(L_4); break;
		case 0x8: TRI_DIP(L_8); break;
		case 0x5: TRI_DIP(L_5); break;
		case 0x9: TRI_DIP(L_9); break;
		case 0x6: TRI_DIP(L_6); break;
		case 0xa: TRI_DIP(L_a); break;
		case 0xc: TRI_DIP(L_c); break;
		case 0x7: v[1]=v[4]; TETRA_L(); break;
		case 0xd: v[2]=v[4]; TETRA_L(); break;
		case 0xe: v[3]=v[4]; TETRA_L(); break;
		default : Error("unexpected left-handed system in CH5P.");
	    }
}

static void	CH4P(CH3D_T *ch3d,CH3D_U p)
{
	CH3D_U	u,v[4],f[4],e[6];

	for (u=0; u<4; u++) v[u]=MOL[p+u];

	if (JORSSS(v,X,Y,Z,NAME)>0) TETRA_R(); else TETRA_L();
}

			/* SV EV RF LF SCE SCCE ECE ECCE */
static WES	DIHED[3]={{ 0, 1, 1, 0,  2,   2,  1,   1},
			  { 1, 2, 1, 0,  0,   0,  2,   2},
			  { 2, 0, 1, 0,  1,   1,  0,   0}};

static void	CH3P(CH3D_T *ch3d,CH3D_U p)
{
	CH3D_U	u,v[3],f[2],e[3];

	for (u=0; u<3; u++) v[u]=MOL[p+u];

	Set(ch3d,3,v,2,f,3,e,DIHED);
}

static void	VONF3(CH3D_T *ch3d,CH3D_U f,CH3D_U *v)
{
	CH3D_U	e=FE[f];

	if (f==RF[e]) {
	    v[0]=EV[e]; v[1]=SV[e]; e=SCE[e];
	}
	else {
	    v[0]=SV[e]; v[1]=EV[e]; e=ECE[e];
	}
	v[2]=(f==RF[e])?SV[e]:EV[e];
}

static int	ITLEFT(CH3D_U v0,CH3D_U v1,CH3D_U v2,
		       CH3D_I *x,CH3D_I *y,CH3D_U *name)
{
	CH3D_U	v[3];
	int	change,sign;

	v[0]=v0; v[1]=v1; v[2]=v2; change=SORTBN(3,v,name,NULL);

/* 0 */	if (!(sign=  JSD1SS(v[0],v[1],v[2],x,y)))
/* 1 */	if (!(sign= -JSD1S (v[1],v[2],y)))
/* 2 */	if (!(sign=  JSD1S (v[0],v[2],y)))
/* 3 */	if (!(sign=  JSD1S (v[1],v[2],x))) sign=(-1);

	return change*sign;
}

static CH3D_U	SHADOW(CH3D_T *ch3d,CH3D_U lmp)
{
	CH3D_U	v[3],e1,e2,f1,f2,v2,rmp=lmp,v1=lmp;
	int	o1,o2;

	do {
	    e1=e2=VE[v1];
	    VONF3(ch3d,f1=(v1==SV[e1])?RF[e1]:LF[e1],v);
	    o1=ITLEFT(v[0],v[1],v[2],X,Y,NAME);
	    for (;;) {
		VONF3(ch3d,f2=(v1==SV[e2])?LF[e2]:RF[e2],v);
		o2=ITLEFT(v[0],v[1],v[2],X,Y,NAME);
		if (o1>0 && o2<0) break;

		if ((e2=(v1==SV[e2])?SCCE[e2]:ECCE[e2])==e1)
		    Error("no silhouette edge in SHADOW.");

		f1=f2; o1=o2;
	    }
	    if ((v2=SV[e2])==v1) v2=EV[e2];

	    CNXV[v1]=v2; CCNXV[v2]=v1; if (LARGER(v2,rmp,X,NAME)) rmp=v2;
	} while ((v1=v2)!=lmp);

	return rmp;
}

#define NOT_YET_JUDGED	0
#define NEWLY_CREATED	1
#define TO_BE_DELETED	2
#define BOUNDARY_EDGE	3

static void	CCLEAR(CH3D_T *ch3d,CH3D_U v0)
{
	CH3D_U	e,e0,f,v=v0;

	do {
	    e=e0=VE[v]; f=(v==SV[e])?RF[e]:LF[e];
	    do {
		ECOLO[e]=FCOLO[f]=NOT_YET_JUDGED;
		if (v==SV[e]) {
		    f=LF[e]; e=SCCE[e];
		}
		else {
		    f=RF[e]; e=ECCE[e];
		}
	    } while (e!=e0);
	} while ((v=HULNV[v])!=v0);
}

static void	EONF3(CH3D_T *ch3d,CH3D_U f,CH3D_U *e)
{
	e[0]=FE[f]; e[1]=(f==RF[e[0]])?SCE[e[0]]:ECE[e[0]];
		    e[2]=(f==RF[e[1]])?SCE[e[1]]:ECE[e[1]];
}

static void	Delete(CH3D_T *ch3d,CH3D_U f0,CH3D_U v0)
{
	CH3D_U	f,f1,list[3],u,e,f2,v,v1,v2,e1,e2,Nf=0,Ne=0,Nv=0;

	FCOLO[FQUE[Nf++]=f0]=TO_BE_DELETED;

	for (f=0; f<Nf; f++) {
	    EONF3(ch3d,f1=FQUE[f],list);
	    for (u=0; u<3; u++) {
		if (ECOLO[e=list[u]]!=NOT_YET_JUDGED) continue;

		if (Ne>=PARA3) Error("edge stack overflow.");

		ECOLO[ESTAC[Ne++]=e]=TO_BE_DELETED;

		if (FCOLO[f2=(f1==RF[e])?LF[e]:RF[e]]==TO_BE_DELETED) continue;

		if (Nf>=PARA2) Error("face stack overflow.");

		FCOLO[FQUE[Nf++]=f2]=TO_BE_DELETED;
	    }
	}
	if (v0!=CH3D_LIM) {
/* begin DELALL */
	    v=HULNV[v0]; do VSTAC[Nv++]=v; while ((v=HULNV[v])!=v0);
/* end DELALL */
	}
	else {
/* begin DELETE */
	    v1=v2=SV[FE[f0]];
	    do {
		e1=e2=VE[v2];
		while (ECOLO[e2]==TO_BE_DELETED)
		    if ((e2=(v2==SV[e2])?SCCE[e2]:ECCE[e2])==e1) {
			VSTAC[Nv++]=v2; break;
		    }
	    } while ((v2=HULNV[v2])!=v1);
/* end DELETE */
	}
	for (u=0; u<Nf; u++) {
/* begin KFBACK */
	    FE[FQUE[u]]=PARA4; PARA4=FQUE[u];
/* end KFBACK */
	}
	for (u=0; u<Ne; u++) {
/* begin KEBACK */
	    SV[ESTAC[u]]=PARA5; PARA5=ESTAC[u];
/* end KEBACK */
	}
	for (u=0; u<Nv; u++) {
	    v=VSTAC[u]; v1=HULNV[v]; v2=HULPV[v]; HULNV[v2]=v1; HULPV[v1]=v2;
	}
}

static void	MERGCH(CH3D_T *ch3d,CH3D_U vL,CH3D_U vR)
{
	CH3D_U	v,eT,e1,v1,e2,v2,e,e0,list[4],f1,f2,f;
	int	move,o1,o2,left;

	vL=SHADOW(ch3d,vL); (void)SHADOW(ch3d,vR);

/* begin COMTAN */
	do {
	    for (move=0; ITLEFT(vR,vL,v=CCNXV[vL],X,Y,NAME)<0; move=1) vL=v;
	    for (      ; ITLEFT(vL,vR,v= CNXV[vR],X,Y,NAME)>0; move=1) vR=v;
	} while (move);
/* end COMTAN */

/* begin GCYLIN */
	CCLEAR(ch3d,vL);
	CCLEAR(ch3d,vR);

	ECOLO[eT=KNEWE(ch3d)]=NEWLY_CREATED; e1=VE[SV[eT]=v1=vL];
					     e2=VE[EV[eT]=v2=vR]; e=eT;
	do {
	    e0=e;

	    list[3]=v2;
	    if (v1==SV[e=e1]) {
		f1=LF[e]; f2=RF[e];
	    }
	    else {
		f1=RF[e]; f2=LF[e];
	    }
	    VONF3(ch3d,f1,list); o1=JORSSS(list,X,Y,Z,NAME); e1=e;
	    for (;;) {
		VONF3(ch3d,f2,list); o2=JORSSS(list,X,Y,Z,NAME);
		if (o1>0 && o2<0) break;

		if ((e1=(v1==SV[e1])?SCE[e1]:ECE[e1])==e)
		    Error("no silhouette edge on left hull.");

		f1=f2; f2=(v1==SV[e1])?RF[e1]:LF[e1]; o1=o2;
	    }
	    list[3]=v1;
	    if (v2==SV[e=e2]) {
		f1=RF[e]; f2=LF[e];
	    }
	    else {
		f1=LF[e]; f2=RF[e];
	    }
	    VONF3(ch3d,f1,list); o1=JORSSS(list,X,Y,Z,NAME); e2=e;
	    for (;;) {
		VONF3(ch3d,f2,list); o2=JORSSS(list,X,Y,Z,NAME);
		if (o1>0 && o2<0) break;

		if ((e2=(v2==SV[e2])?SCCE[e2]:ECCE[e2])==e)
		    Error("no silhouette edge on right hull.");

		f1=f2; f2=(v2==SV[e2])?LF[e2]:RF[e2]; o1=o2;
	    }
	    list[0]=v2;
	    list[1]=v1;
	    list[2]=(v2==SV[e2])?EV[e2]:SV[e2];
	    list[3]=(v1==SV[e1])?EV[e1]:SV[e1];
	    if (JORSSS(list,X,Y,Z,NAME)<0) {
		left=0; ECOLO[e2]=BOUNDARY_EDGE; v2=list[2];
	    }
	    else {
		left=1; ECOLO[e1]=BOUNDARY_EDGE; v1=list[3];
	    }
	    FCOLO[f=KNEWF(ch3d)]=NEWLY_CREATED;
	    if (v1==vL && v2==vR)
		e=eT;
	    else
		ECOLO[e=KNEWE(ch3d)]=NEWLY_CREATED;

	    FE[LF[e]=RF[e0]=f]=e; SV[e]=v1; EV[e]=v2;
	    if (left)
		SCCE[ECCE[e0]=e]=SCE[ECE[e]=e0]=e1;
	    else
		ECE[SCE[e0]=e]=ECCE[SCCE[e]=e0]=e2;
	} while (e!=eT);

	e=eT;
	for (;;) {
	    if ((e=SCE[e])==eT) {
		Delete(ch3d,RF[VE[vL]],vL); break;
	    }
	    if (ECOLO[e]==BOUNDARY_EDGE) {
		Delete(ch3d,(SV[e]==vL)?LF[e]:RF[e],CH3D_LIM); break;
	    }
	}
	e=eT;
	for (;;) {
	    if ((e=ECCE[e])==eT) {
		Delete(ch3d,RF[VE[vR]],vR); break;
	    }
	    if (ECOLO[e]==BOUNDARY_EDGE) {
		Delete(ch3d,(SV[e]==vR)?RF[e]:LF[e],CH3D_LIM); break;
	    }
	}
	e1=eT;
	do {
	    VE[v1=SV[e1]]=VE[v2=EV[e1]]=e1;
	    if (ECOLO[SCE[e1]]==NEWLY_CREATED) {
		e2=SCE[e1]; e=ECCE[e1];
		if (SV[e]==v2) {
		    SCE[e]=e1; ECCE[e]=e2; RF[e]=RF[e1];
		}
		else {
		    ECE[e]=e1; SCCE[e]=e2; LF[e]=RF[e1];
		}
	    }
	    else {
		e2=ECCE[e1]; e=SCE[e1];
		if (SV[e]==v1) {
		    ECE[e]=e2; SCCE[e]=e1; LF[e]=RF[e1];
		}
		else {
		    SCE[e]=e2; ECCE[e]=e1; RF[e]=RF[e1];
		}
	    }
	} while ((e1=e2)!=eT);

	v1=HULNV[vL];
	v2=HULNV[vR]; HULPV[HULNV[vL]=v2]=vL; HULPV[HULNV[vR]=v1]=vR;
/* end GCYLIN */
}

static CH3D_U	Store(CH3D_U N,CH3D_U w,CH3D_U P,CH3D_U *w1,CH3D_U *w2)
{
	CH3D_U	u;

	for (u=0; u<P; u++) if (w1[u]==w) w2[N++]=u;

	return N;
}

static void	Restore(CH3D_U N,CH3D_U P,CH3D_U *w1,CH3D_U *w2)
{
	CH3D_U	u;

	for (u=0; u<P; u++) w1[u]=CH3D_LIM;
	for (u=0; u<N; u++) w1[w2[u]]=u;
	while (u<P) w2[u++]=CH3D_LIM;
}

void	CH3D(CH3D_T *ch3d,CH3D_U Np,int name)
{
	CH3D_U	u,p,n,nofc,list[3];

	if (Np<4 || Np>PARA1) Error("bad number of points.");

	for (u=0; u<Np; u++) MOL[u]=u;

	if (name) {
	    Sort_NAME(ch3d,1,Np);

	    for (p=NAME[MOL[0]], u=1; u<Np; u++, p=n)
		if (p==(n=NAME[MOL[u]])) Error("bad value of \"NAME\".");
	}
	else
	    for (u=0; u<Np; u++) NAME[u]=u;

	Sort_X(ch3d,1,Np);

/* begin KINIT */
	PARA4=0; for (FE[u=PARA2-1]=CH3D_LIM; u>0; u--) FE[u-1]=u;
	PARA5=0; for (SV[u=PARA3-1]=CH3D_LIM; u>0; u--) SV[u-1]=u;
/* end KINIT */

	if (Np==5) {
	    CH5P(ch3d); p=MOL[0];
	}
	else {
	    nofc=(Np-(n=3-(Np-1)%4)*3)/4; p=0;
	    for (u=0; u<nofc; u++) {
		CH4P(ch3d,p); CLIST[u]=MOL[p]; p+=4;
	    }
	    for (nofc+=n; u<nofc; u++) {
		CH3P(ch3d,p); CLIST[u]=MOL[p]; p+=3;
	    }
	    while (nofc>1) {
		for (u=1; u<nofc; u+=2) {
		    MERGCH(ch3d,p=CLIST[u-1],CLIST[u]); CLIST[u/2]=p;
		}
		if (nofc%2==0)
		    nofc/=2;
		else {
		    CLIST[u=nofc/2]=CLIST[nofc-1]; nofc=u+1;
		}
	    }
	    p=CLIST[0];
	}
	for (u=0; u<PARA1; u++) WV1[u]=0;
	u=p; do WV1[u]=1; while ((u=HULNV[u])!=p);

	for (u=0; u<PARA3; u++) WE1[u]=1;
	for (u=PARA5; u!=CH3D_LIM; u=SV[u]) WE1[u]=0;

	for (u=0; u<PARA2; u++) WF1[u]=1;
	for (u=PARA4; u!=CH3D_LIM; u=FE[u]) WF1[u]=0;

	for (u=0; u<PARA2; u++)
	    if (WF1[u]==1) {
		VONF3(ch3d,u,list);
		if (JSD1SS(list[0],list[1],list[2],X,Y) ||
		    JSD1SS(list[0],list[1],list[2],Z,X) ||
		    JSD1SS(list[0],list[1],list[2],Y,Z)) {
		    WF1[u]=3;
		    WV1[list[0]]|=2; WV1[list[1]]|=2; WV1[list[2]]|=2;
		    EONF3(ch3d,u,list);
		    WE1[list[0]]|=2; WE1[list[1]]|=2; WE1[list[2]]|=2;
		}
	    }

	Restore(ch3d->Nv2=Store(Store(ch3d->Nv1=Store(0,3,PARA1,WV1,WV2),
							2,PARA1,WV1,WV2),
							1,PARA1,WV1,WV2),
							  PARA1,WV1,WV2);
	Restore(ch3d->Ne2=Store(Store(ch3d->Ne1=Store(0,3,PARA3,WE1,WE2),
							2,PARA3,WE1,WE2),
							1,PARA3,WE1,WE2),
							  PARA3,WE1,WE2);
	Restore(ch3d->Nf2=Store(Store(ch3d->Nf1=Store(0,3,PARA2,WF1,WF2),
							2,PARA2,WF1,WF2),
							1,PARA2,WF1,WF2),
							  PARA2,WF1,WF2);
}

void	CH3D_e2p(CH3D_T *ch3d,CH3D_U e,CH3D_U p[/* 2 */])
{
	e=WE2[e]; p[0]=SV[e]; p[1]=EV[e];
}

void	CH3D_f2p(CH3D_T *ch3d,CH3D_U f,CH3D_U p[/* 3 */])
{
	VONF3(ch3d,WF2[f],p);
}

void	CH3D_f2e(CH3D_T *ch3d,CH3D_U f,CH3D_U e[/* 3 */])
{
	EONF3(ch3d,WF2[f],e);
/*
	e[0]=WE1[e[0]]; e[1]=WE1[e[1]]; e[2]=WE1[e[2]];
*/
}

int	CH3D_TL(CH3D_U v0,CH3D_U v1,CH3D_U v2,CH3D_I *x,CH3D_I *y,CH3D_U *name)
{
	return (name==NULL)?JSD1SS(v0,v1,v2,x,y):ITLEFT(v0,v1,v2,x,y,name);
}

int	CH3D_OR(CH3D_U v[/* 4 */],CH3D_I *x,CH3D_I *y,CH3D_I *z,CH3D_U *name)
{
	return (name==NULL)?JD1SSS(v,x,y,z):JORSSS(v,x,y,z,name);
}
