
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include "cell.h"
#include "csi.h"
#include "ch3d.h"

extern void	Error(char *msg),
		RangeList(char *str,size_t lim,char *flag),
		ReadSliceImage(char *path,int Nx,int Ny,Cell **cell);

static int	Append(int i2,int *X,int *Y,int x,int y)
{
	int	i1,x1,y1,x2=X[i2]-x,y2=Y[i2]-y;

	while ((i1=i2)>0) {
	    --i2; x1=x2; y1=y2; if ((x2=X[i2]-x)*y1<(y2=Y[i2]-y)*x1) break;
	}
	++i1; X[i1]=x; Y[i1]=y++; x1=0; y1=(-1);
	while ((i2=i1)>0) {
	    --i1; x2=x1; y2=y1; if ((x1=X[i1]-x)*y2<(y1=Y[i1]-y)*x2) break;
	}
	++i2; X[i2]=x; Y[i2]=y; return i2;
}

static double	*U,*V;

static int	Compare(CH3D_U *I1,CH3D_U *I2)
{
	return (U[*I1]<U[*I2])?-1:(U[*I1]==U[*I2])?
	       (V[*I1]<V[*I2])?-1:(V[*I1]==V[*I2])?0:1:1;
}

static CH3D_U	Half(long long h,long long t,long long s,CH3D_U *I,CH3D_U *J)
{
	CH3D_U	n,i,j;
	double	u0,v0,u1,v1,u2,v2;

	n=J[0]=I[h   ]; u0=U[n]; v0=V[n];
	n=J[1]=I[h+=s]; u1=U[n]; v1=V[n]; n=2;
	while ((h+=s)!=t) {
	    i=I[h]; u2=U[i]; v2=V[i];
	    while ((u1-u0)*(v2-v0)>=(v1-v0)*(u2-u0)) {
		u1=u0; v1=v0;
		if (--n>=2) {
		    j=J[n-2]; u0=U[j]; v0=V[j];
		}
		else
		    break;
	    }
	    u0=u1; v0=v1; u1=u2; v1=v2; J[n++]=i;
	}
	return n;
}

static double	Square(double d)
{
	return d*d;
}

/*
#define EPS	1e-9
*/
#define EPS	1e-7	/* to avoid error when compiled by "icc -fast" */

static double	A1,B1,xo1,yo1,zo1,xa1,ya1,za1,xb1,yb1,zb1,xc1,yc1,zc1,
		A2,B2,xo2,yo2,zo2,xa2,ya2,za2,xb2,yb2,zb2,xc2,yc2,zc2;

static void	CH_MM(CH3D_U N,
		      CH3D_I *X,CH3D_I *Y,CH3D_I *Z,
		      CH3D_U *V2P,CH3D_U *ALL,CH3D_U *CH,
		      double cx,double cy,double cz,double c1,double c2)
{
	double	vz,vy,ux,vx,uy,u1,v1,tu,u3,u2,tv,v3,v2,u0,du,v0,dv,w,
		bu,bv,au,av,b0,a0,b1,b2,a1,a2,a3,b3,uo,vo,ua,va,ub,vb;
	CH3D_U	i,j,p,n,o,i1,j1,i2,j2;

#define uz	0.0

	if ((vz= -sqrt(1.0-cz*cz))< -EPS) {
	    vy=(ux= -cy/vz)*cz; vx= -(uy=cx/vz)*cz;
	}
	else {
	    ux=vy=1.0; uy=vx=0.0;
	}
	for (i=N, j=0; j<N; ALL[j++]=i++) {
	    p=V2P[j]; U[i]=ux*(double)X[p]+uy*(double)Y[p]+uz*(double)Z[p];
		      V[i]=vx*(double)X[p]+vy*(double)Y[p]+vz*(double)Z[p];
	}
	qsort(ALL,(size_t)N,sizeof(CH3D_U),(int (*)())Compare);

	n=Half(0,N,1,ALL,CH)-2;
	for (i=1; i<=n; i++) CH[i-1]=CH[i];

	n+=Half(N-1,-1,-1,ALL,CH+n);
	for (i=0; i<n; i++) {
	    j=CH[i]; U[i]=U[j]; V[i]=V[j];
	}
	o=n-1; u1=U[o];
	       v1=V[o]; tu=(u3=U[1])-(u2=U[0]);
			tv=(v3=V[1])-(v2=V[0]); j=1; B2=0.0;
	for (i=0; i<n; i++) {
	    u0=u1; du=(u1=U[i])-u0;
	    v0=v1; dv=(v1=V[i])-v0;
	    while (dv*tu-du*tv>0.0) {
		j=(j+1)%n; u2=u3; tu=(u3=U[j])-u2;
			   v2=v3; tv=(v3=V[j])-v2;
	    }
	    w=Square(dv*(u0-=u2)-du*(v0-=v2))/(Square(du)+Square(dv));
	    if (i==0 || w<A1) {
		A1=w; i1=i; j1=j;
	    }
	    if ((w=Square(u0)+Square(v0))>B2) {
		B2=w; i2=i; j2=j;
	    }
	    if ((w=Square(u1-u2)+Square(v1-v2))>B2) {
		B2=w; i2=i+1; j2=j;
	    }
	}
	u0=U[i1]; v0=V[i1]; i1=(i1+o)%n;
	w=sqrt(Square(tu=u0-U[i1])+Square(tv=v0-V[i1])); bu=tu/w; bv=tv/w;

	j1=(j1+o)%n; u1=U[j1]; v1=V[j1];

	i2=(i2+o)%n; u2=U[i2]; v2=V[i2];
	j2=(j2+o)%n; u3=U[j2]; v3=V[j2];
	w=sqrt(Square(du=u3-u2)+Square(dv=v3-v2)); au=dv/w; av= -du/w;

	for (i=0; i<n; i++) {
	    b0=bu*U[i]+bv*V[i]; a0=au*U[i]+av*V[i];
	    if (i==0) {
		b1=b2=b0; a1=a2=a0;
	    }
	    else {
		if (b0<b1) b1=b0; else if (b0>b2) b2=b0;
		if (a0<a1) a1=a0; else if (a0>a2) a2=a0;
	    }
	}
	A1=sqrt(A1); B1=b2-b1; a0=  bv*u0-bu*v0; a3=  bv*u1-bu*v1;
	B2=sqrt(B2); A2=a2-a1; b0= -av*u2+au*v2; b3= -av*u3+au*v3;

	uo=bv*a0+bu*b1; vo= -bu*a0+bv*b1;
	ua=bv*a3+bu*b1; va= -bu*a3+bv*b1;
	ub=bv*a0+bu*b2; vb= -bu*a0+bv*b2;
	xo1=uo*ux+vo*vx+c1*cx; yo1=uo*uy+vo*vy+c1*cy; zo1=uo*uz+vo*vz+c1*cz;
	xa1=ua*ux+va*vx+c1*cx; ya1=ua*uy+va*vy+c1*cy; za1=ua*uz+va*vz+c1*cz;
	xb1=ub*ux+vb*vx+c1*cx; yb1=ub*uy+vb*vy+c1*cy; zb1=ub*uz+vb*vz+c1*cz;
	xc1=uo*ux+vo*vx+c2*cx; yc1=uo*uy+vo*vy+c2*cy; zc1=uo*uz+vo*vz+c2*cz;

	uo=au*a1-av*b0; vo=av*a1+au*b0;
	ua=au*a2-av*b0; va=av*a2+au*b0;
	ub=au*a1-av*b3; vb=av*a1+au*b3;
	xo2=uo*ux+vo*vx+c1*cx; yo2=uo*uy+vo*vy+c1*cy; zo2=uo*uz+vo*vz+c1*cz;
	xa2=ua*ux+va*vx+c1*cx; ya2=ua*uy+va*vy+c1*cy; za2=ua*uz+va*vz+c1*cz;
	xb2=ub*ux+vb*vx+c1*cx; yb2=ub*uy+vb*vy+c1*cy; zb2=ub*uz+vb*vz+c1*cz;
	xc2=uo*ux+vo*vx+c2*cx; yc2=uo*uy+vo*vy+c2*cy; zc2=uo*uz+vo*vz+c2*cz;
}

#define PV	(Cell)(~0)

typedef struct LIST {
		int    		I1,I2,*X1,*Y1,*X2,*Y2,z;
		struct LIST	*next;
	} List;

typedef struct {
		CH3D_U	n;
		List	*list,**next;
	} Table;

#define MALLOC(var,type,noe)	(var=(type *)malloc(sizeof(type)*(noe)))==NULL
#define REALLOC_INT(var,noe)	\
	(var=(int *)realloc(var,sizeof(int)*(noe)))==NULL

int	main(int argc,char **argv)
{
	char			**path,*mae="memory allocation error.";
	int			Nx,Ny,Nz,y,z,w,x,x1,x2,y1,y2,z1,z2;
	static char		flag[PV+1];
	Cell			**cell,**last;
	unsigned long		c,n1=0,n2=0;
	static int		X1[PV+1],X2[PV+1];
	static List		*L[PV+1];
	static Table		T[PV+1];
	static unsigned long	S1[PV+1],S2[PV+1];
	List			*l,*next;
	CH3D_U			n,*V2P,*ALL,*CH,P[3],i,p,j;
	CH3D_T			*ch3d;
	CH3D_I			*X,*Y,*Z;
	double			dx,dy,dz,d,D,d1,d2,C,cx,cy,cz,c1,c2;

	if (argc!=4) Error("usage : si_bbs directory nameFile rangeList");

	path=CheckSliceImages(argv[1],argv[2],&Nx,&Ny,&Nz,NULL);

	RangeList(argv[3],PV,flag);

	if (MALLOC(cell,Cell *,Ny) ||
	    MALLOC(*cell,Cell,Ny*Nx) ||
	    MALLOC(last,Cell *,Ny) ||
	    MALLOC(*last,Cell,Ny*Nx)) Error(mae);

	for (y=1; y<Ny; y++) {
	    cell[y]=cell[y-1]+Nx; last[y]=last[y-1]+Nx;
	}
	for (c=0; c<=PV; c++) {
	    X1[c]=Nx; L[c]=NULL; T[c].n=0; T[c].next=(&(T[c].list));
	}
	for (z=0; z<=Nz; z++) {
	    if (z==Nz)
		w=1;
	    else {
		ReadSliceImage(path[z],Nx,Ny,cell); free(path[z]); w=z==0;
	    }
	    for (y=0; y<Ny; y++) {
		for (x=0; x<Nx; x++) {
		    if (w) {
			if (flag[c=cell[y][x]]) {
			    X2[c]=x; if (X1[c]==Nx) X1[S1[n1++]=c]=x;
			}
		    }
		    else if (last[y][x]!=cell[y][x]) {
			if (flag[c=last[y][x]]) {
			    X2[c]=x; if (X1[c]==Nx) X1[S1[n1++]=c]=x;
			}
			if (flag[c=cell[y][x]]) {
			    X2[c]=x; if (X1[c]==Nx) X1[S1[n1++]=c]=x;
			}
		    }
		    last[y][x]=cell[y][x];
		}
		while (n1>0) {
		    if ((l=L[c=S1[--n1]])==NULL) {
			if (MALLOC(l=L[c],List,1) ||
			    MALLOC(l->X1,int,Ny+1-y) ||
			    MALLOC(l->Y1,int,Ny+1-y) ||
			    MALLOC(l->X2,int,Ny+1-y) ||
			    MALLOC(l->Y2,int,Ny+1-y)) Error(mae);

			l->X1[0]=l->X1[1]=  X1[c]; l->Y1[0]=l->Y2[0]=y  ;
			l->X2[0]=l->X2[1]= -X2[c]; l->Y1[1]=l->Y2[1]=y+1;
			l->I1=l->I2=1; l->z=z; S2[n2++]=c;
		    }
		    else {
			l->I1=Append(l->I1,l->X1,l->Y1, X1[c],y);
			l->I2=Append(l->I2,l->X2,l->Y2,-X2[c],y);
		    }
		    X1[c]=Nx;
		}
	    }
	    while (n2>0) {
		l=L[c=S2[--n2]];
		if (REALLOC_INT(l->X1,l->I1+1) ||
		    REALLOC_INT(l->Y1,l->I1+1) ||
		    REALLOC_INT(l->X2,l->I2+1) ||
		    REALLOC_INT(l->Y2,l->I2+1)) Error(mae);

		T[c].n+=(l->I1+l->I2+2);
		*(T[c].next)=l; *(T[c].next=(&(l->next)))=L[c]=NULL;
	    }
	}
	free(*last); free(last); free(*cell); free(cell); free(path);

	n=0; for (c=0; c<=PV; c++) if (T[c].n>n) n=T[c].n;

	if (n!=0)
{
	ch3d=InitCH3D(n); X=CH3D_X(ch3d); Y=CH3D_Y(ch3d); Z=CH3D_Z(ch3d);
			  U=(double *)ch3d->we1; V=(double *)ch3d->we2;
			  V2P=ch3d->wv2; ALL=ch3d->wv5; CH=ch3d->wv6;

	for (c=0; c<=PV; c++) if (T[c].n!=0)
{
	(void)printf("%lu\n",c);

	n=0;
	for (l=T[c].list; l!=NULL; l=next) {
	    for (w=0; w<=l->I1; w++) {
		X[n]=  l->X1[w]; Y[n]=l->Y1[w]; Z[n++]=l->z;
	    }
	    for (w=l->I2; w>=0; w--) {
		X[n]=1-l->X2[w]; Y[n]=l->Y2[w]; Z[n++]=l->z;
	    }
	    free(l->Y2); free(l->X2);
	    free(l->Y1); free(l->X1); next=l->next; free(l);
	}
	CH3D(ch3d,n,0);

	for (n=0; n<ch3d->Nf1; n++) {
	    CH3D_f2p(ch3d,n,P);
	    x=X[P[0]]; x1=X[P[1]]-x; x2=X[P[2]]-x;
	    y=Y[P[0]]; y1=Y[P[1]]-y; y2=Y[P[2]]-y;
	    z=Z[P[0]]; z1=Z[P[1]]-z; z2=Z[P[2]]-z;
	    d=sqrt(Square(dx=(double)(-y1*z2+y2*z1))+
		   Square(dy=(double)(-z1*x2+z2*x1))+
		   Square(dz=(double)(-x1*y2+x2*y1))); dx/=d; dy/=d; dz/=d;
	    d1=d2=dx*(double)x+dy*(double)y+dz*(double)z;
	    for (i=0; i<ch3d->Nv1; i++) {
		p=CH3D_V2P(ch3d,i);
		if ((d=dx*(double)X[p]+dy*(double)Y[p]+dz*(double)Z[p])>d2)
		    d2=d;
	    }
	    D=d2-d1;
	    if (n==0 || D<C) {
		C=D; cx=dx; cy=dy; cz=dz; c1=d1; c2=d2;
	    }
	}
	CH_MM(ch3d->Nv1,X,Y,Z,V2P,ALL,CH,cx,cy,cz,c1,c2);

#define PRINTF(A,B,C,xo,yo,zo,xa,ya,za,xb,yb,zb,xc,yc,zc)	\
	(void)printf("%lg\t%lg\t%lg\n" \
		     "%lg\t%lg\t%lg\t" \
		     "%lg\t%lg\t%lg\t" \
		     "%lg\t%lg\t%lg\t" \
		     "%lg\t%lg\t%lg\n", \
		     A,B,C,xo,yo,zo,xa,ya,za,xb,yb,zb,xc,yc,zc)

	PRINTF(C,A1,B1,xo1,yo1,zo1,xc1,yc1,zc1,xa1,ya1,za1,xb1,yb1,zb1);
	PRINTF(C,A2,B2,xo2,yo2,zo2,xc2,yc2,zc2,xa2,ya2,za2,xb2,yb2,zb2);

	C=0.0;
	for (i=0; i<ch3d->Nv1-1; i++) {
	    p=CH3D_V2P(ch3d,i); x=X[p]; y=Y[p]; z=Z[p];
	    for (j=i+1; j<ch3d->Nv1; j++) {
		p=CH3D_V2P(ch3d,j);
		if ((D=Square(dx=(double)(X[p]-x))+
		       Square(dy=(double)(Y[p]-y))+
		       Square(dz=(double)(Z[p]-z)))>C) {
		    C=D; cx=dx; x1=x; x2=X[p];
			 cy=dy; y1=y; y2=Y[p];
			 cz=dz; z1=z; z2=Z[p];
		}
	    }
	}
	C=sqrt(C); cx/=C; cy/=C; cz/=C;

	CH_MM(ch3d->Nv1,X,Y,Z,V2P,ALL,CH,
	      cx,cy,cz,cx*(double)x1+cy*(double)y1+cz*(double)z1,
		       cx*(double)x2+cy*(double)y2+cz*(double)z2);

	PRINTF(A1,B1,C,xo1,yo1,zo1,xa1,ya1,za1,xb1,yb1,zb1,xc1,yc1,zc1);
	PRINTF(A2,B2,C,xo2,yo2,zo2,xa2,ya2,za2,xb2,yb2,zb2,xc2,yc2,zc2);
}
	TermCH3D(ch3d);
}
	return 0;
}
