
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include "cell.h"
#include "csi.h"

#define EPS	1e-9

extern void	Error(char *msg),
		RangeList(char *str,size_t lim,char *obj),
		ReadSliceImage(char *path,int Nx,int Ny,Cell **slice),
		chull3_(int *V,int *X,int *Y,int *Z,int *name,int *para,
			int *ve,int *fe,int *sv,int *ev,int *rf,int *lf,
			int *sce,int *scce,int *ece,int *ecce,
			int *wv1,int *wv2,int *wv3,int *wv4,int *wv5,int *wv6,
			int *we1,int *we2,int *wf1,int *wf2);

static int	Append(int i2,int *X,int *Y,int x,int y)
{
	int	i1,x1,y1,x2=X[i2]-x,y2=Y[i2]-y;

	while ((i1=i2)>0) {
	    --i2; x1=x2; y1=y2; if ((x2=X[i2]-x)*y1<(y2=Y[i2]-y)*x1) break;
	}
	++i1; X[i1]=x; Y[i1]=y++; x1=0; y1=(-1);
	while ((i2=i1)>0) {
	    --i1; x2=x1; y2=y1; if ((x1=X[i1]-x)*y2<(y1=Y[i1]-y)*x2) break;
	}
	++i2; X[i2]=x; Y[i2]=y; return i2;
}

static double	*P,*Q;

static int	Compare(int *I1,int*I2)
{
	double	p1,p2,q1,q2;

	return ((p1=P[*I1])<(p2=P[*I2]))?-1:(p1==p2)?
	       ((q1=Q[*I1])<(q2=Q[*I2]))?-1:(q1==q2)?0:1:1;
}

static int	Half(int h,int t,int s,int*I,int *J)
{
	int	n,i,j;
	double	p0,q0,p1,q1,p2,q2;

	n=J[0]=I[h   ]; p0=P[n]; q0=Q[n];
	n=J[1]=I[h+=s]; p1=P[n]; q1=Q[n]; n=2;
	while ((h+=s)!=t) {
	    i=I[h]; p2=P[i]; q2=Q[i];
	    while ((p1-p0)*(q2-q0)>=(q1-q0)*(p2-p0)) {
		p1=p0; q1=q0;
		if (--n>=2) {
		    j=J[n-2]; p0=P[j]; q0=Q[j];
		}
		else
		    break;
	    }
	    p0=p1; q0=q1; p1=p2; q1=q2; J[n++]=i;
	}
	return n;
}

static double	Square(double d)
{
	return d*d;
}

static int	*I,*J;
static double	A1,B1,xo1,yo1,zo1,xa1,ya1,za1,xb1,yb1,zb1,xc1,yc1,zc1,
		A2,B2,xo2,yo2,zo2,xa2,ya2,za2,xb2,yb2,zb2,xc2,yc2,zc2;

static void	CH_MM(int V,int *X,int *Y,int *Z,int *name,
		      double cx,double cy,double cz,double c1,double c2)
{
	double	qz,qy,px,qx,py,p1,q1,tp,p3,p2,tq,q3,q2,p0,dp,q0,dq,w,
		bp,bq,ap,aq,b0,a0,b1,b2,a1,a2,a3,b3,po,qo,pa,qa,pb,qb;
	int	i,j,k,n,o,i1,j1,i2,j2;

	if ((qz= -sqrt(1.0-cz*cz))< -EPS) {
	    qy=(px= -cy/qz)*cz; qx= -(py=cx/qz)*cz;
	}
	else {
	    px=qy=1.0; py=qx=0.0;
	}

#define pz	0.0

	for (i=V, j=0; j<V; j++, i++) {
	    k=name[j]; P[i]=px*(double)X[k]+py*(double)Y[k]+pz*(double)Z[k];
		       Q[i]=qx*(double)X[k]+qy*(double)Y[k]+qz*(double)Z[k];
	    I[j]=i;
	}
	qsort(I,(size_t)V,sizeof(int),(int (*)())Compare);

	n=Half(0,V,1,I,J)-2;
	for (i=1; i<=n; i++) J[i-1]=J[i];

	n+=Half(V-1,-1,-1,I,J+n);
	for (i=0; i<n; i++) {
	    j=J[i]; P[i]=P[j]; Q[i]=Q[j];
	}
	o=n-1; p1=P[o];
	       q1=Q[o]; tp=(p3=P[1])-(p2=P[0]);
			tq=(q3=Q[1])-(q2=Q[0]); j=1; B2=0.0;
	for (i=0; i<n; i++) {
	    p0=p1; dp=(p1=P[i])-p0;
	    q0=q1; dq=(q1=Q[i])-q0;
	    while (dq*tp-dp*tq>0) {
		j=(j+1)%n; p2=p3; tp=(p3=P[j])-p2;
			   q2=q3; tq=(q3=Q[j])-q2;
	    }

#define S(p,q)	(Square(p)+Square(q))

	    w=Square(dq*(p0-=p2)-dp*(q0-=q2))/S(dp,dq);
	    if (i==0 || w<A1) {
		A1=w; i1=i; j1=j;
	    }
	    if ((w=S(p0,q0))>B2) {
		B2=w; i2=i; j2=j;
	    }
	    if ((w=S(p1-p2,q1-q2))>B2) {
		B2=w; i2=i+1; j2=j;
	    }
	}
	p0=P[i1]; q0=Q[i1]; i1=(i1+o)%n;
	w=sqrt(S(tp=p0-P[i1],tq=q0-Q[i1])); bp=tp/w; bq=tq/w;

	j1=(j1+o)%n; p1=P[j1]; q1=Q[j1];

	i2=(i2+o)%n; p2=P[i2]; q2=Q[i2];
	j2=(j2+o)%n; p3=P[j2]; q3=Q[j2];
	w=sqrt(S(dp=p3-p2,dq=q3-q2)); ap=dq/w; aq= -dp/w;

	for (i=0; i<n; i++) {
	    b0=bp*P[i]+bq*Q[i]; a0=ap*P[i]+aq*Q[i];
	    if (i==0) {
		b1=b2=b0; a1=a2=a0;
	    }
	    else {
		if (b0<b1) b1=b0; else if (b0>b2) b2=b0;
		if (a0<a1) a1=a0; else if (a0>a2) a2=a0;
	    }
	}
	A1=sqrt(A1); B1=b2-b1; a0=  bq*p0-bp*q0; a3=  bq*p1-bp*q1;
	B2=sqrt(B2); A2=a2-a1; b0= -aq*p2+ap*q2; b3= -aq*p3+ap*q3;

	po=bq*a0+bp*b1; qo= -bp*a0+bq*b1;
	pa=bq*a3+bp*b1; qa= -bp*a3+bq*b1;
	pb=bq*a0+bp*b2; qb= -bp*a0+bq*b2;
	xo1=po*px+qo*qx+c1*cx; yo1=po*py+qo*qy+c1*cy; zo1=po*pz+qo*qz+c1*cz;
	xa1=pa*px+qa*qx+c1*cx; ya1=pa*py+qa*qy+c1*cy; za1=pa*pz+qa*qz+c1*cz;
	xb1=pb*px+qb*qx+c1*cx; yb1=pb*py+qb*qy+c1*cy; zb1=pb*pz+qb*qz+c1*cz;
	xc1=po*px+qo*qx+c2*cx; yc1=po*py+qo*qy+c2*cy; zc1=po*pz+qo*qz+c2*cz;

	po=ap*a1-aq*b0; qo=aq*a1+ap*b0;
	pa=ap*a2-aq*b0; qa=aq*a2+ap*b0;
	pb=ap*a1-aq*b3; qb=aq*a1+ap*b3;
	xo2=po*px+qo*qx+c1*cx; yo2=po*py+qo*qy+c1*cy; zo2=po*pz+qo*qz+c1*cz;
	xa2=pa*px+qa*qx+c1*cx; ya2=pa*py+qa*qy+c1*cy; za2=pa*pz+qa*qz+c1*cz;
	xb2=pb*px+qb*qx+c1*cx; yb2=pb*py+qb*qy+c1*cy; zb2=pb*pz+qb*qz+c1*cz;
	xc2=po*px+qo*qx+c2*cx; yc2=po*py+qo*qy+c2*cy; zc2=po*pz+qo*qz+c2*cz;
}

#define PV	(Cell)(~0)

typedef struct LIST {
		int    		I1,I2,*X1,*Y1,*X2,*Y2,z;
		struct LIST	*next;
	} List;

typedef struct {
		int	V;
		List	*list,**ptr;
	} Table;

int	main(int argc,char **argv)
{
	char		**path;
	int		Nx,Ny,Nz,y,n1,n2,i,z,x,j,V,k,*X,*Y,*Z,
			*name,*ve,*fe,*sv,*ev,*rf,*lf,*sce,*scce,*ece,*ecce,
			*wv1,*wv2,*wv3,*wv4,*wv5,*wv6,*we1,*we2,*wf1,*wf2,
			x1,x2,para[10],F,v1,v2,v3,x3,y3,y1,y2,z3,z1,z2,v;
	static char	object[PV+1];
	Cell		**cell,**last,c;
	static int	X1[PV+1],X2[PV+1];
	static List	*L[PV+1];
	static Table	T[PV+1];
	static Cell	S1[PV+1],S2[PV+1];
	List		*l;
	long long	lx,ly,lz;
	double		dx,dy,dz,d,D,d1,d3,d2,C,cx,cy,cz,c1,c2;

	if (argc!=4) Error("usage : si_bbs directory nameFile rangeList");

	path=CheckSliceImages(argv[1],argv[2],&Nx,&Ny,&Nz,NULL);

	RangeList(argv[3],PV,object);

	if ((cell=(Cell **)malloc(sizeof(Cell *)*Ny))==NULL ||
	    (*cell=(Cell *)malloc(sizeof(Cell)*Ny*Nx))==NULL ||
	    (last=(Cell **)malloc(sizeof(Cell *)*Ny))==NULL ||
	    (*last=(Cell *)malloc(sizeof(Cell)*Ny*Nx))==NULL)
	  Error("no allocatable memory for image.");

	for (y=1; y<Ny; y++) {
	  cell[y]=cell[y-1]+Nx; last[y]=last[y-1]+Nx;
	}
	n1=n2=0;
	for (i=0; i<=PV; i++) {
	  X1[i]=Nx; L[i]=NULL; T[i].V=0; T[i].ptr=(&(T[i].list));
	}
	for (z=0; z<=Nz; z++) {
	  if (z<Nz) {
	    ReadSliceImage(path[z],Nx,Ny,cell); free(path[z]);
	  }
	  for (y=0; y<Ny; y++) {
	    for (x=0; x<Nx; x++) {
	      if (z>0 && object[c=last[y][x]]) {
		X2[c]=x; if (X1[c]==Nx) X1[S1[n1++]=c]=x;
	      }
	      if (z<Nz && object[c=last[y][x]=cell[y][x]]) {
		X2[c]=x; if (X1[c]==Nx) X1[S1[n1++]=c]=x;
	      }
	    }
	    while (n1>0) {
	      if ((l=L[c=S1[--n1]])==NULL) {
		if ((l=L[c]=(List *)malloc(sizeof(List)))==NULL ||
		    (l->X1=(int *)malloc(sizeof(int)*(i=Ny+1-y)))==NULL ||
		    (l->Y1=(int *)malloc(sizeof(int)*i))==NULL ||
		    (l->X2=(int *)malloc(sizeof(int)*i))==NULL ||
		    (l->Y2=(int *)malloc(sizeof(int)*i))==NULL)
		  Error("no allocatable memory for 2D convex hull.");

		l->I1=l->I2=1;
		l->X1[0]=l->X1[1]=  X1[c]; l->Y1[0]=l->Y2[0]=y;
		l->X2[0]=l->X2[1]= -X2[c]; l->Y1[1]=l->Y2[1]=y+1;

		S2[n2++]=c;
	      }
	      else {
		l->I1=Append(l->I1,l->X1,l->Y1, X1[c],y);
		l->I2=Append(l->I2,l->X2,l->Y2,-X2[c],y);
	      }
	      X1[c]=Nx;
	    }
	  }
	  while (n2>0) {
	    l=L[c=S2[--n2]];
	    if ((l->X1=(int *)realloc(l->X1,sizeof(int)*(i=l->I1+1)))==NULL ||
		(l->Y1=(int *)realloc(l->Y1,sizeof(int)*i))==NULL ||
		(l->X2=(int *)realloc(l->X2,sizeof(int)*(j=l->I2+1)))==NULL ||
		(l->Y2=(int *)realloc(l->Y2,sizeof(int)*j))==NULL)
	      Error("memory re-allocation error.");

	    l->z=z; l->next=L[c]=NULL;

	    T[c].V+=(i+j); *(T[c].ptr)=l; T[c].ptr=(&(l->next));
	  }
	}
	free(path); free(*last); free(last); free(*cell); free(cell);

	V=0; for (i=0; i<=PV; i++) if (T[i].V>V) V=T[i].V;

	if (V!=0) {
	  i=V*2+1; j=i*3; k=i*4;
	  if ((X=(int *)malloc(sizeof(int)*i))==NULL ||
	      (Y=(int *)malloc(sizeof(int)*i))==NULL ||
	      (Z=(int *)malloc(sizeof(int)*i))==NULL ||
	      (name=(int *)malloc(sizeof(int)*i))==NULL ||
	      (ve=(int *)malloc(sizeof(int)*i))==NULL ||
	      (fe=(int *)malloc(sizeof(int)*j))==NULL ||
	      (sv=(int *)malloc(sizeof(int)*k))==NULL ||
	      (ev=(int *)malloc(sizeof(int)*k))==NULL ||
	      (rf=(int *)malloc(sizeof(int)*k))==NULL ||
	      (lf=(int *)malloc(sizeof(int)*k))==NULL ||
	      (sce=(int *)malloc(sizeof(int)*k))==NULL ||
	      (scce=(int *)malloc(sizeof(int)*k))==NULL ||
	      (ece=(int *)malloc(sizeof(int)*k))==NULL ||
	      (ecce=(int *)malloc(sizeof(int)*k))==NULL ||
	      (wv1=(int *)malloc(sizeof(int)*i))==NULL ||
	      (wv2=(int *)malloc(sizeof(int)*i))==NULL ||
	      (wv3=(int *)malloc(sizeof(int)*i))==NULL ||
	      (wv4=(int *)malloc(sizeof(int)*i))==NULL ||
	      (wv5=(int *)malloc(sizeof(int)*i))==NULL ||
	      (wv6=(int *)malloc(sizeof(int)*i))==NULL ||
	      (we1=(int *)malloc(sizeof(int)*k))==NULL ||
	      (we2=(int *)malloc(sizeof(int)*k))==NULL ||
	      (wf1=(int *)malloc(sizeof(int)*j))==NULL ||
	      (wf2=(int *)malloc(sizeof(int)*j))==NULL)
	    Error("no allocatable memory for 3D convex hull.");

	  P=(double *)we1; Q=(double *)we2; J=(I=wf2)+V;

	  for (i=0; i<=PV; i++)
	    if (T[i].V!=0) {
	      (void)printf("%d\n",i);

	      V=0;
	      for (l=T[i].list; l!=NULL; l=l->next) {
	        z=l->z;
		for (x1=x2=0; x1<=l->I1 || x2<=l->I2; V++) {
		  if (x2>l->I2 || (x1<=l->I1 && l->Y1[x1]<=l->Y2[x2])) {
		    X[V]=  l->X1[x1]; Y[V]=l->Y1[x1]; ++x1;
		  }
		  else {
		    X[V]=1-l->X2[x2]; Y[V]=l->Y2[x2]; ++x2;
		  }
		  Z[name[V]=V]=z;
		}
	      }
	      para[0]=n1=V+1; chull3_(&V,X,Y,Z,name,para,
				      ve,fe,sv,ev,rf,lf,sce,scce,ece,ecce,
				      wv1,wv2,wv3,wv4,wv5,wv6,we1,we2,wf1,wf2);

	      V=0; for (j=0; j<n1; j++) if (wv5[j]) name[V++]=j; 

	      n2=n1*3; F=0;
	      for (j=0; j<n2; j++)
		if (wf1[j]) {
		  if (j==rf[k=fe[j]-1]-1) {
		    v1=ev[k]-1; v2=sv[k]-1; k=sce[k]-1;
		  }
		  else {
		    v1=sv[k]-1; v2=ev[k]-1; k=ece[k]-1;
		  }
		  v3=(j==rf[k]-1)?sv[k]-1:ev[k]-1;

		  x3=X[v3]; x1=x3-X[v1]; x2=x3-X[v2];
		  y3=Y[v3]; y1=y3-Y[v1]; y2=y3-Y[v2];
		  z3=Z[v3]; z1=z3-Z[v1]; z2=z3-Z[v2];
		  lx=(long long)y1*(long long)z2-(long long)z1*(long long)y2;
		  ly=(long long)z1*(long long)x2-(long long)x1*(long long)z2;
		  lz=(long long)x1*(long long)y2-(long long)y1*(long long)x2;
		  if (lx!=0 || ly!=0 || lz!=0) {
		    d=sqrt(Square(dx=(double)lx)+
			   Square(dy=(double)ly)+
			   Square(dz=(double)lz)); dx/=d; dy/=d; dz/=d;

		    d1=dx*(double)x3+dy*(double)y3+dz*(double)z3; D=0.0;
		    for (k=0; k<V; k++) {
		      v=name[k];
		      if ((d=fabs((d3=dx*(double)X[v]
				     +dy*(double)Y[v]
				     +dz*(double)Z[v])-d1))>D) {
			D=d; d2=d3;
		      }
		    }
		    if (++F==1 || D<C) {
		      C=D; cx=dx; cy=dy; cz=dz; c1=d1; c2=d2;
		    }
		  }
		}

	      CH_MM(V,X,Y,Z,name,cx,cy,cz,c1,c2);

#define P(A,B,C,xo,yo,zo,xa,ya,za,xb,yb,zb,xc,yc,zc)	\
	(void)printf("%lg\t%lg\t%lg\n" \
		     "%lg\t%lg\t%lg\t" \
		     "%lg\t%lg\t%lg\t" \
		     "%lg\t%lg\t%lg\t" \
		     "%lg\t%lg\t%lg\n", \
		     A,B,C,xo,yo,zo,xa,ya,za,xb,yb,zb,xc,yc,zc)

	      P(C,A1,B1,xo1,yo1,zo1,xc1,yc1,zc1,xa1,ya1,za1,xb1,yb1,zb1);
	      P(C,A2,B2,xo2,yo2,zo2,xc2,yc2,zc2,xa2,ya2,za2,xb2,yb2,zb2);

	      n1=V-1; C=0.0;
	      for (j=0; j<n1; j++) {
		v=name[j]; x=X[v]; y=Y[v]; z=Z[v];
		for (k=j+1; k<V; k++) {
		  v=name[k];
		  if ((D=Square(dx=(double)((x3=X[v])-x))+
			 Square(dy=(double)((y3=Y[v])-y))+
			 Square(dz=(double)((z3=Z[v])-z)))>C) {
		    C=D; cx=dx; x1=x; x2=x3;
			 cy=dy; y1=y; y2=y3;
			 cz=dz; z1=z; z2=z3;
		  }
		}
	      }
	      C=sqrt(C); cx/=C; cy/=C; cz/=C;

	      CH_MM(V,X,Y,Z,name,cx,cy,cz,
		    cx*(double)x1+cy*(double)y1+cz*(double)z1,
		    cx*(double)x2+cy*(double)y2+cz*(double)z2);

	      P(A1,B1,C,xo1,yo1,zo1,xa1,ya1,za1,xb1,yb1,zb1,xc1,yc1,zc1);
	      P(A2,B2,C,xo2,yo2,zo2,xa2,ya2,za2,xb2,yb2,zb2,xc2,yc2,zc2);
	    }
	}
	return 0;
}
