
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "cell.h"
#include "csi.h"
#include "ch3d.h"
#include "sif_cm.h"

extern void	Error(char *msg),
		RangeList(char *str,size_t lim,char *flag),
		ReadSliceImage(char *path,int Nx,int Ny,Cell **cell);

#define LABEL	unsigned long
#define FORMAT	"%lu"

static LABEL	Trace(LABEL *link,LABEL l1)
{
	LABEL	l2=link[l1];

	return (l2==0)?l1:(link[l1]=Trace(link,l2));
}

static int	Append(int i2,int *X,int *Y,int x,int y)
{
	int	i1,x1,y1,x2=X[i2]-x,y2=Y[i2]-y;

	while ((i1=i2)>0) {
	    --i2; x1=x2; y1=y2; if ((x2=X[i2]-x)*y1<(y2=Y[i2]-y)*x1) break;
	}
	++i1; X[i1]=x; Y[i1]=y++; x1=0; y1=(-1);
	while ((i2=i1)>0) {
	    --i1; x2=x1; y2=y1; if ((x1=X[i1]-x)*y2<(y1=Y[i1]-y)*x2) break;
	}
	++i2; X[i2]=x; Y[i2]=y; return i2;
}

#define PV	(Cell)(~0)
#define LEN	1024
#define INC	4096
#define EPS	1e-9

typedef	struct {
		int	n,*x,*y;
	} CH2D_T;

#define MALLOC(var,type,noe)	(var=(type *)malloc(sizeof(type)*(noe)))
#define REALLOC(var,type,noe)	(var=(type *)realloc(var,sizeof(type)*(noe)))

int	main(int argc,char **argv)
{
	char		**paths,flag[(size_t)PV+1],path[LEN],
			*mae="memory allocation error.";
	int		Nx,Ny,Nz,*X1,*Y1,*X2,*Y2,y,x,z,i,x1,x2,I1,I2,*Xz,*Yz,
			**X3,**X4,y0,y1,y2,z0,z1,z2,y3,y4,z3,z4,x0,**X0;
	LABEL		**last,*link,L,l,lxy,Nc=0,c3=0,c2=0,c1=0;
	Cell		**cell;
	CH2D_T		*ch2d;
	CH3D_U		n,P[3];
	CH3D_T		*ch3d;
	CH3D_I		*X,*Y,*Z;
	double		gy,gz,bx,z5,y5,x3,x4,x5;
	unsigned short	R[4]={65535,65535,0,0},
			G[4]={65535,0,65535,0},
			B[4]={65535,0,0,65535};

	if (argc!=4 && argc!=5)
	    Error("usage : si_ch orgDir nameFile rangeList {newDir}");

	paths=CheckSliceImages(argv[1],argv[2],&Nx,&Ny,&Nz,NULL);

	RangeList(argv[3],PV,flag);

	if (MALLOC(cell,Cell *,Ny)==NULL ||
	    MALLOC(*cell,Cell,Ny*Nx)==NULL ||
	    MALLOC(last,LABEL *,Ny+1)==NULL ||
	    MALLOC(*last,LABEL,(Ny+1)*(Nx+1))==NULL ||
	    MALLOC(link,LABEL,INC)==NULL ||
	    MALLOC(ch2d,CH2D_T,Nz+1)==NULL ||
	    MALLOC(X1,int,Ny+1)==NULL ||
	    MALLOC(Y1,int,Ny+1)==NULL ||
	    MALLOC(X2,int,Ny+1)==NULL ||
	    MALLOC(Y2,int,Ny+1)==NULL) Error(mae);

	for (y=1; y<Ny; y++) cell[y]=cell[y-1]+Nx;

	++last; ++last[-1];
	for (x=0; x<Nx; x++) last[-1][x]=1;
	for (y=0; y<Ny; y++) {
	    last[y]=last[y-1]+Nx+1; last[y][-1]=1;
	}
	link[1]=0;

	L=2; for (y=0; y<Ny; y++)
	     for (x=0; x<Nx; x++) last[y][x]=1;

	n=0;
	for (z=0; z<=Nz; z++) {
	  if (z==Nz)
	    for (y=0; y<Ny; y++)
	    for (x=0; x<Nx; x++) cell[y][x]=2;
	  else {
	    ReadSliceImage(paths[z],Nx,Ny,cell);

	    for (y=0; y<Ny; y++)
	    for (x=0; x<Nx; x++) cell[y][x]=flag[cell[y][x]]!=0;
	  }
	  i=0;
	  for (y=0; y<Ny; y++) {
	    x1=Nx;
	    for (x=0; x<Nx; x++)
	      if ((last[y][x]==0)+(cell[y][x]==1)==1 && (x2=x)<x1) x1=x;

	    if (x1!=Nx)
	      if (i==0) {
		X1[0]=X1[1]=  x1; Y1[0]=Y2[0]=y  ;
		X2[0]=X2[1]= -x2; Y1[1]=Y2[1]=y+1; I1=I2=i=1;
	    }
	    else {
	      I1=Append(I1,X1,Y1, x1,y);
	      I2=Append(I2,X2,Y2,-x2,y);
	    }
	  }
	  if (i==0)
	    ch2d[z].n=0;
	  else {
	    n+=(i=ch2d[z].n=I1+I2+2);
	    if (MALLOC(Xz=ch2d[z].x,int,i)==NULL ||
		MALLOC(Yz=ch2d[z].y,int,i)==NULL) Error(mae);

	    for (i=0; i<=I2; i++) {
	      *(Xz++)=1-X2[i]; *(Yz++)=Y2[i];
	    }
	    for (i=I1; i>=0; i--) {
	      *(Xz++)=  X1[i]; *(Yz++)=Y1[i];
	    }
	  }
	  for (y=0; y<Ny; y++) {
	    for (x=0; x<Nx; x++) {
	      if (cell[y][x]==1)
		l=0;
	      else if (cell[y][x]==2) {
		if ((l=last[y][x])!=0 && (l=Trace(link,l))!=1) l=link[l]=1;
	      }
	      else if ((l=last[y][x])!=0) {
		l=Trace(link,l);
		if ((lxy=last[y-1][x])!=0 && (lxy=Trace(link,lxy))!=l)
		  if (lxy==1) l=link[l]=1; else link[lxy]=l;

		if ((lxy=last[y][x-1])!=0 && (lxy=Trace(link,lxy))!=l)
		  if (lxy==1) l=link[l]=1; else link[lxy]=l;
	      }
	      else if ((l=last[y-1][x])!=0) {
		l=Trace(link,l);
		if ((lxy=last[y][x-1])!=0 && (lxy=Trace(link,lxy))!=l)
		  if (lxy==1) l=link[l]=1; else link[lxy]=l;
	      }
	      else if ((l=last[y][x-1])!=0)
		l=Trace(link,l);
	      else if (L%INC!=0 || REALLOC(link,LABEL,L+INC)!=NULL)
		link[l=L++]=0;
	      else
		Error(mae);

	      last[y][x]=l;
	    }
	    if (z<Nz && l>1) link[l]=1;
	  }
	  if (z<Nz) {
	    y=Ny-1;
	    for (x=0; x<Nx; x++)
	      if ((l=last[y][x])!=0 && (l=Trace(link,l))!=1) link[l]=1;
	  }
	}
	free(Y2); free(X2); free(Y1); free(X1);

	if (REALLOC(link,LABEL,L)==NULL) Error(mae);

	ch3d=InitCH3D(n); X=CH3D_X(ch3d); Y=CH3D_Y(ch3d); Z=CH3D_Z(ch3d);

	n=0;
	for (z=0; z<=Nz; z++)
	    if ((i=ch2d[z].n)!=0) {
		while (--i>=0) {
		    X[n]=ch2d[z].x[i]; Y[n]=ch2d[z].y[i]; Z[n++]=z;
		}
		free(ch2d[z].y); free(ch2d[z].x);
	    }

        free(ch2d);

	if (MALLOC(X3,int *,Nz)==NULL ||
	    MALLOC(*X3,int,Nz*Ny)==NULL ||
	    MALLOC(X4,int *,Nz)==NULL ||
	    MALLOC(*X4,int,Nz*Ny)==NULL) Error(mae);

	CH3D(ch3d,n,0);

	for (z=0; z<Nz; z++) {
	    if (z>0) {
		X3[z]=X3[z-1]+Ny; X4[z]=X4[z-1]+Ny;
	    }
	    for (y=0; y<Ny; y++) {
		X3[z][y]=Nx; X4[z][y]=(-1);
	    }
	}
	for (n=0; n<ch3d->Nf1; n++) {
	    CH3D_f2p(ch3d,n,P);
	    y0=Y[P[0]]; y1=Y[P[1]]; y2=Y[P[2]];
	    z0=Z[P[0]]; z1=Z[P[1]]; z2=Z[P[2]];
	    if ((x=(y1-y0)*(z2-z0)-(y2-y0)*(z1-z0))!=0) {
		if (y1<y2) {
		    y3=(y0<y1)?y0:y1; y4=(y0>y2)?y0:y2;
		}
		else {
		    y3=(y0<y2)?y0:y2; y4=(y0>y1)?y0:y1;
		}
		if (z1<z2) {
		    z3=(z0<z1)?z0:z1; z4=(z0>z2)?z0:z2;
		}
		else {
		    z3=(z0<z2)?z0:z2; z4=(z0>z1)?z0:z1;
		}
		y1-=y0; y2-=y0;
		z1-=z0; z2-=z0; x0=X[P[0]]; x1=X[P[1]]-x0; x2=X[P[2]]-x0;
		gy=(double)(z2*x1-z1*x2)/(double)x;
		gz=(double)(x2*y1-x1*y2)/(double)x;
		if (x<0) {
		    X0=X3; bx=(double)x0+0.5-EPS;
		}
		else {
		    X0=X4; bx=(double)x0+0.5+EPS;
		}
		for (z=z3; z<z4; z++) {
		    z5=(double)(z-z0)+0.5;
		    for (y=y3; y<y4; y++) {
			y5=(double)(y-y0)+0.5;
			x5=(x3=y5*(double)z2-(double)y2*z5)+
			   (x4=(double)y1*z5-y5*(double)z1);
			if ((x3>=0.0 && x4>=0.0 && x5<=(double)x) ||
			    (x3<=0.0 && x4<=0.0 && x5>=(double)x))
			    X0[z][y]=(int)(gz*z5+gy*y5+bx);
		    }
		}
	    }
	}
	TermCH3D(ch3d);

	if (argc==5) for (i=0; argv[1][i++]!='\0'; ) ;

	for (l=2; l<L; l++) if (link[l]!=0) (void)Trace(link,l);
	for (l=2; l<L; l++)
	    if (link[l]==0) {
		link[l]=l; ++Nc;
	    }

	L=2; for (y=0; y<Ny; y++)
	     for (x=0; x<Nx; x++) last[y][x]=1;

	for (z=0; z<Nz; z++) {
	    ReadSliceImage(paths[z],Nx,Ny,cell);

	    for (y=0; y<Ny; y++) {
		x1=X3[z][y]; x2=X4[z][y];
		for (x=0; x<Nx; x++)
		    if (flag[cell[y][x]]) {
			last[y][x]=0; cell[y][x]=3; ++c3;
		    }
		    else {
			if ((l=last[y  ][x  ])==0 &&
			    (l=last[y-1][x  ])==0 &&
			    (l=last[y  ][x-1])==0) l=link[L++];

			if ((last[y][x]=l)!=1) {
			    cell[y][x]=2; ++c2;
			}
			else if (x1<=x && x<x2) {
			    cell[y][x]=1; ++c1;
			}
			else
			    cell[y][x]=0;
		    }
	    }
	    if (argc==5) {
		(void)sprintf(path,"%s/%s",argv[4],paths[z]+i);
		StoreImageFile_CM(path,Nx,Ny,2,R,G,B,cell,NULL);
	    }
	    free(paths[z]);
	}
	(void)printf(FORMAT "\n"
		     FORMAT "\t" FORMAT "\n"
		     FORMAT "\n",c1,c2,Nc,c3);

	free(*X4); free(X4); free(*X3); free(X3); free(link);
	free(last[-1]-1); free(last-1); free(*cell); free(cell); return 0;
}
