
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "cell.h"
#include "csi.h"
#include "sif_cm.h"
#include "sif.h"

#define EPS	1e-9	/* ==DBL_EPSILON */

#define LABEL	unsigned long long

#ifdef	__MSVCRT__
#define FORMAT	"%I64u"
#else
#define FORMAT	"%llu"
#endif

static unsigned short	cmR[4]={65535,65535,0,0},
			cmG[4]={65535,0,65535,0},
			cmB[4]={65535,0,0,65535};

extern void	Error(),RangeList(),ReadSliceImage(),chull3_();

static LABEL	Trace(link,l1)
LABEL		*link,l1;
{
	LABEL	l2=link[l1];

	return (l2==0)?l1:(link[l1]=Trace(link,l2));
}

static int	Append(i2,X,Y,x,y)
int		i2,*X,*Y,x,y;
{
	int	i1,x1,y1,x2=X[i2]-x,y2=Y[i2]-y;

	while ((i1=i2)>0) {
	    --i2; x1=x2; y1=y2; if ((x2=X[i2]-x)*y1<(y2=Y[i2]-y)*x1) break;
	}
	++i1; X[i1]=x; Y[i1]=y++; x1=0; y1=(-1);
	while ((i2=i1)>0) {
	    --i1; x2=x1; y2=y1; if ((x1=X[i1]-x)*y2<(y1=Y[i1]-y)*x2) break;
	}
	++i2; X[i2]=x; Y[i2]=y; return i2;
}

static int	FV3(X,Y,Z,fe,sv,ev,rf,sce,ece,f,v1,v2,v3)
int		*X,*Y,*Z,*fe,*sv,*ev,*rf,*sce,*ece,f,*v1,*v2,*v3;
{
	int		e=fe[f]-1;
	long long	x1,y1,z1,x2,y2,z2;

	if (f==rf[e]-1) {
	    *v1=ev[e]-1; *v2=sv[e]-1; e=sce[e]-1;
	}
	else {
	    *v1=sv[e]-1; *v2=ev[e]-1; e=ece[e]-1;
	}
	*v3=(f==rf[e]-1)?sv[e]-1:ev[e]-1;

	return (x1=X[*v1]-X[*v3])*(y2=Y[*v2]-Y[*v3])!=
	       (y1=Y[*v1]-Y[*v3])*(x2=X[*v2]-X[*v3]) ||
	       (z1=Z[*v1]-Z[*v3])*x2!=x1*(z2=Z[*v2]-Z[*v3]) || y1*z2!=z1*y2;
}

#define PV	(Cell)(~0)
#define LEN	1024

int	main(argc,argv)
int	argc;
char	**argv;
{
	char	**paths,pv[(size_t)PV+1],path[LEN];
	int	Nx,Ny,Nz,i,j,K,*X1,*Y1,*X2,*Y2,*X,*Y,*Z,y,x,V,z,x1,x2,I1,I2,
		N1,N2,*name,*ve,*fe,*sv,*ev,*rf,*lf,*sce,*scce,*ece,*ecce,
		*wv1,*wv2,*wv3,*wv4,*wv5,*wv6,*we1,*we2,*wf1,*wf2,para[10],
		**X3,**X4,v1,v2,v3,y0,y1,y2,z0,z1,z2,y3,y4,z3,z4,x0,**X5;
	LABEL	**last,*link,L,l,lxy,CL=0,C3=0,C2=0,C1=0;
	Cell	**cell;
	double	gy,gz,bx,z5,y5,x3,x4,x5;

	if (argc!=4 && argc!=5)
	  Error("usage : si_ch orgDir nameFile rangeList {newDir}");

	paths=CheckSliceImages(argv[1],argv[2],&Nx,&Ny,&Nz,NULL);

	RangeList(argv[3],(size_t)PV,pv);

	i=Ny+1; j=i*2*(Nz+1); K=Nx+1;

	if ((cell=(Cell **)malloc(sizeof(Cell *)*Ny))==NULL ||
	    (*cell=(Cell *)malloc(sizeof(Cell)*Ny*Nx))==NULL ||
	    (last=(LABEL **)malloc(sizeof(LABEL *)*i))==NULL ||
	    (*last=(LABEL *)malloc(sizeof(LABEL)*i*K))==NULL ||
	    (X1=(int *)malloc(sizeof(int)*i))==NULL ||
	    (Y1=(int *)malloc(sizeof(int)*i))==NULL ||
	    (X2=(int *)malloc(sizeof(int)*i))==NULL ||
	    (Y2=(int *)malloc(sizeof(int)*i))==NULL ||
	    (X=(int *)malloc(sizeof(int)*j))==NULL ||
	    (Y=(int *)malloc(sizeof(int)*j))==NULL ||
	    (Z=(int *)malloc(sizeof(int)*j))==NULL ||
	    (link=(LABEL *)malloc(sizeof(LABEL)*K))==NULL)
	  Error("no allocatable memory for sampling points.");

	for (y=1; y<Ny; y++) cell[y]=cell[y-1]+Nx;

	++last; ++last[-1];
	for (x=0; x<Nx; x++) last[-1][x]=1;
	for (y=0; y<Ny; y++) {
	  last[y]=last[y-1]+K; last[y][-1]=1;
	}
	link[1]=0;

	L=2; for (y=0; y<Ny; y++)
	     for (x=0; x<Nx; x++) last[y][x]=1;

	V=0;
	for (z=0; z<Nz; z++) {
	  ReadSliceImage(paths[z],Nx,Ny,cell);

	  i=0;
	  for (y=0; y<Ny; y++) {
	    x1=Nx;
	    for (x=0; x<Nx; x++)
	      if (pv[cell[y][x]]) {
		last[y][x]=0; if ((x2=x)<x1) x1=x;
	      }
	      else if ((l=last[y][x])!=0) {
		l=Trace(link,l);
		if ((lxy=last[y-1][x])!=0 && (lxy=Trace(link,lxy))!=l)
		  if (lxy==1) l=link[l]=1; else link[lxy]=l;

		if ((lxy=last[y][x-1])!=0 && (lxy=Trace(link,lxy))!=l)
		  if (lxy==1) l=link[l]=1; else link[lxy]=l;

		last[y][x]=l;
	      }
	      else {
		if ((l=last[y-1][x])!=0) {
		  l=Trace(link,l);
		  if ((lxy=last[y][x-1])!=0 && (lxy=Trace(link,lxy))!=l)
		    if (lxy==1) l=link[l]=1; else link[lxy]=l;

		  last[y][x]=l;
		}
		else if ((l=last[y][x-1])!=0)
		  last[y][x]=Trace(link,l);
		else {
		  if (L%K==0 &&
		     (link=(LABEL *)realloc(link,sizeof(LABEL)*(L+K)))==NULL)
		    Error("no re-allocatable memory for sampling points.");

		  link[last[y][x]=L++]=0;
		}
		if ((x2=x)<x1) x1=x;
	      }

	    if (x1!=Nx)
	      if (i==0) {
		X1[0]=X1[1]=  x1; Y1[0]=Y2[0]=y  ;
		X2[0]=X2[1]= -x2; Y1[1]=Y2[1]=y+1; I1=I2=i=1;
	      }
	      else {
		I1=Append(I1,X1,Y1, x1,y);
		I2=Append(I2,X2,Y2,-x2,y);

		if (V!=0 && Nx>1 &&
		   (l=last[y][Nx-1])!=0 && (l=Trace(link,l))!=1) link[l]=1;
	      }
	  }
	  if (i!=0) {
	    if (V!=0 && Ny>1) {
	      i=Ny-1;
	      for (x=Nx-1; x>0; x--)
		if ((l=last[i][x])!=0 && (l=Trace(link,l))!=1) link[l]=1;
	    }
	    for (x1=x2=0; x1<=I1 || x2<=I2; Z[V++]=z)
	      if (x2>I2 || (x1<=I1 && Y1[x1]<=Y2[x2])) {
		X[V]=  X1[x1]; Y[V]=Y1[x1]; ++x1;
	      }
	      else {
		X[V]=1-X2[x2]; Y[V]=Y2[x2]; ++x2;
	      }
	  }
	}
	i=0;
	for (y=0; y<Ny; y++) {
	  x1=Nx;
	  for (x=0; x<Nx; x++)
	    if (last[y][x]==0 && (x2=x)<x1) x1=x;

	  if (x1!=Nx)
	    if (i==0) {
	      X1[0]=X1[1]=  x1; Y1[0]=Y2[0]=y  ;
	      X2[0]=X2[1]= -x2; Y1[1]=Y2[1]=y+1; I1=I2=i=1;
	    }
	    else {
	      I1=Append(I1,X1,Y1, x1,y);
	      I2=Append(I2,X2,Y2,-x2,y);
	    }
	}
	if (i!=0)
	  for (x1=x2=0; x1<=I1 || x2<=I2; Z[V++]=z)
	    if (x2>I2 || (x1<=I1 && Y1[x1]<=Y2[x2])) {
	      X[V]=  X1[x1]; Y[V]=Y1[x1]; ++x1;
	    }
	    else {
	      X[V]=1-X2[x2]; Y[V]=Y2[x2]; ++x2;
	    }

	free(Y2); free(X2); free(Y1); free(X1);

	if (argc==5) {
	  for (K=0; argv[1][K]!='\0'; K++) ; ++K;
	}
	if (V!=0) {
	  for (y=Ny-1; y>0; y--)
	  for (x=Nx-1; x>0; x--)
	    if ((l=last[y][x])!=0 && (l=Trace(link,l))!=1) link[l]=1;

	  N1=V+1; N2=N1*3; i=N1*4;

	  if ((link=(LABEL *)realloc(link,sizeof(LABEL)*L))==NULL ||
	      (X=(int *)realloc(X,sizeof(int)*N1))==NULL ||
	      (Y=(int *)realloc(Y,sizeof(int)*N1))==NULL ||
	      (Z=(int *)realloc(Z,sizeof(int)*N1))==NULL ||
	      (name=(int *)malloc(sizeof(int)*N1))==NULL ||
	      (ve=(int *)malloc(sizeof(int)*N1))==NULL ||
	      (fe=(int *)malloc(sizeof(int)*N2))==NULL ||
	      (sv=(int *)malloc(sizeof(int)*i))==NULL ||
	      (ev=(int *)malloc(sizeof(int)*i))==NULL ||
	      (rf=(int *)malloc(sizeof(int)*i))==NULL ||
	      (lf=(int *)malloc(sizeof(int)*i))==NULL ||
	      (sce=(int *)malloc(sizeof(int)*i))==NULL ||
	      (scce=(int *)malloc(sizeof(int)*i))==NULL ||
	      (ece=(int *)malloc(sizeof(int)*i))==NULL ||
	      (ecce=(int *)malloc(sizeof(int)*i))==NULL ||
	      (wv1=(int *)malloc(sizeof(int)*N1))==NULL ||
	      (wv2=(int *)malloc(sizeof(int)*N1))==NULL ||
	      (wv3=(int *)malloc(sizeof(int)*N1))==NULL ||
	      (wv4=(int *)malloc(sizeof(int)*N1))==NULL ||
	      (wv5=(int *)malloc(sizeof(int)*N1))==NULL ||
	      (wv6=(int *)malloc(sizeof(int)*N1))==NULL ||
	      (we1=(int *)malloc(sizeof(int)*i))==NULL ||
	      (we2=(int *)malloc(sizeof(int)*i))==NULL ||
	      (wf1=(int *)malloc(sizeof(int)*N2))==NULL ||
	      (wf2=(int *)malloc(sizeof(int)*N2))==NULL)
	    Error("no allocatable memory for convex hull.");

	  for (i=0; i<V; i++) name[i]=i;

	  para[0]=N1; chull3_(&V,X,Y,Z,name,para,
			      ve,fe,sv,ev,rf,lf,sce,scce,ece,ecce,
			      wv1,wv2,wv3,wv4,wv5,wv6,we1,we2,wf1,wf2);

	  free(wf2); free(we2); free(we1);
	  free(wv6); free(wv5); free(wv4); free(wv3); free(wv2); free(wv1); 
	  free(ecce); free(scce); free(lf); free(ve); free(name);

	  if ((X3=(int **)malloc(sizeof(int *)*Nz))==NULL ||
	      (X4=(int **)malloc(sizeof(int *)*Nz))==NULL ||
	      (*X3=(int *)malloc(sizeof(int)*Nz*Ny))==NULL ||
	      (*X4=(int *)malloc(sizeof(int)*Nz*Ny))==NULL)
	    Error("no allocatable memory for rasterization.");

	  for (z=0; z<Nz; z++) {
	    if (z!=0) {
	      X3[z]=X3[z-1]+Ny; X4[z]=X4[z-1]+Ny;
	    }
	    for (y=0; y<Ny; y++) {
	      X3[z][y]=Nx; X4[z][y]=(-1);
	    }
	  }
	  for (i=0; i<N2; i++)
	    if (wf1[i] && FV3(X,Y,Z,fe,sv,ev,rf,sce,ece,i,&v1,&v2,&v3)) {
	      y0=Y[v1]; y1=Y[v2]; y2=Y[v3];
	      z0=Z[v1]; z1=Z[v2]; z2=Z[v3];
	      if ((x=(y1-y0)*(z2-z0)-(y2-y0)*(z1-z0))!=0) {
		if (y1<y2) {
		    y3=(y0<y1)?y0:y1; y4=(y0>y2)?y0:y2;
		}
		else {
		    y3=(y0<y2)?y0:y2; y4=(y0>y1)?y0:y1;
		}
		if (z1<z2) {
		    z3=(z0<z1)?z0:z1; z4=(z0>z2)?z0:z2;
		}
		else {
		    z3=(z0<z2)?z0:z2; z4=(z0>z1)?z0:z1;
		}
		y1-=y0; y2-=y0;
		z1-=z0; z2-=z0;
		x0=X[v1]; x1=X[v2]-x0; x2=X[v3]-x0;
		gy=(double)(z2*x1-z1*x2)/(double)x;
		gz=(double)(x2*y1-x1*y2)/(double)x;
		if (x<0) {
		    X5=X3; bx=(double)x0+0.5-EPS;
		}
		else {
		    X5=X4; bx=(double)x0+0.5+EPS;
		}
		for (z=z3; z<z4; z++) {
		  z5=(double)(z-z0)+0.5;
		  for (y=y3; y<y4; y++) {
		    y5=(double)(y-y0)+0.5;
		    x3=y5*(double)z2-(double)y2*z5;
		    x4=(double)y1*z5-y5*(double)z1; x5=x3+x4;
		    if ((x3>=0.0 && x4>=0.0 && x5<=(double)x) ||
			(x3<=0.0 && x4<=0.0 && x5>=(double)x))
		      X5[z][y]=(int)(gz*z5+gy*y5+bx);
		  }
	        }
	      }
	    }

	  for (l=2; l<L; l++) if (link[l]!=0) (void)Trace(link,l);
	  for (l=2; l<L; l++)
	    if (link[l]==0) {
	      link[l]=l; ++CL;
	    }

	  L=2; for (y=0; y<Ny; y++)
	       for (x=0; x<Nx; x++) last[y][x]=1;

	  for (z=0; z<Nz; z++) {
	    ReadSliceImage(paths[z],Nx,Ny,cell);

	    for (y=0; y<Ny; y++) {
	      x1=X3[z][y]; x2=X4[z][y];
	      for (x=0; x<Nx; x++)
		if (pv[cell[y][x]]) {
		  last[y][x]=0; cell[y][x]=3; ++C3;
		}
		else {
		  if ((l=last[y  ][x  ])==0 &&
		      (l=last[y-1][x  ])==0 &&
		      (l=last[y  ][x-1])==0) l=link[L++];

		  if ((last[y][x]=l)!=1) {
		    cell[y][x]=2; ++C2;
		  }
		  else if (x1<=x && x<x2) {
		    cell[y][x]=1; ++C1;
		  }
		  else
		    cell[y][x]=0;
		}
	    }
	    if (argc==5) {
	      (void)sprintf(path,"%s/%s",argv[4],paths[z]+K);
	      StoreImageFile_CM(path,Nx,Ny,2,cmR,cmG,cmB,cell,NULL);
	    }
	  }
	}
	else if (argc==5) {
	  for (y=0; y<Ny; y++)
	  for (x=0; x<Nx; x++) cell[y][x]=0;

	  for (z=0; z<Nz; z++) {
	    (void)sprintf(path,"%s/%s",argv[4],paths[z]+K);
	    StoreImageFile(path,Nx,Ny,1,cell,NULL);
	  }
	}
	(void)printf(FORMAT "\n"
		     FORMAT "\t" FORMAT "\n"
		     FORMAT "\n",C1,C2,CL,C3);

	return 0;
}
