
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "cell.h"
#include "csi.h"
#include "ch3d.h"

extern void	Error(char *msg),
		RangeList(char *str,size_t lim,char *flag),
		ReadSliceImage(char *path,int Nx,int Ny,Cell **cell);

static int	Append(int i2,int *X,int *Y,int x,int y)
{
	int	i1,x1,y1,x2=X[i2]-x,y2=Y[i2]-y;

	while ((i1=i2)>0) {
	    --i2; x1=x2; y1=y2; if ((x2=X[i2]-x)*y1<(y2=Y[i2]-y)*x1) break;
	}
	++i1; X[i1]=x; Y[i1]=y++; x1=0; y1=(-1);
	while ((i2=i1)>0) {
	    --i1; x2=x1; y2=y1; if ((x1=X[i1]-x)*y2<(y1=Y[i1]-y)*x2) break;
	}
	++i2; X[i2]=x; Y[i2]=y; return i2;
}

static void	PutN(FILE *file,char *mem)
{
	(void)fwrite(mem,(size_t)4,(size_t)1,file);
}

static void	PutR(FILE *file,char *mem)
{
	char	c[4];

	c[0]=mem[3]; c[1]=mem[2]; c[2]=mem[1]; c[3]=mem[0]; PutN(file,c);
}

static int	Normal(CH3D_U *P,CH3D_I *X,CH3D_I *Y)
{
	CH3D_U	p1=P[1],p2=P[2];
	CH3D_I	x=X[*P],y=Y[*P];

	return (X[p1]-x)*(Y[p2]-y)!=(X[p2]-x)*(Y[p1]-y);
}

#define PV	(Cell)(~0)

typedef struct {
		int	n,*x,*y;
	} CH2D_T;

#define MALLOC(var,type,noe)	(var=(type *)malloc(sizeof(type)*(noe)))==NULL

int	main(int argc,char **argv)
{
	short		s=1;
	char		**paths,flag[PV+1],*A,*B,*a,*b,*W,
			*mae="memory allocation error.",
			*c=(char *)&s;
	int		Nx,Ny,Nz,*Xl,*Yl,*Xr,*Yr,y,z,x,i,l,r,L,R,*Xz,*Yz;
	CH2D_T		*ch2d;
	Cell		**C;
	CH3D_U		n,p,P[3];
	CH3D_T		*ch3d;
	CH3D_I		*X,*Y,*Z;
	FILE		*file;
	void		(*Put)()=((short)*c==s)?PutN:PutR;
	float		f;
	unsigned	u;

	if (argc!=5)
	    Error("usage : si_ch_zcp directory nameFile rangeList ZCP");

	paths=CheckSliceImages(argv[1],argv[2],&Nx,&Ny,&Nz,NULL);

	RangeList(argv[3],PV,flag);

	if (MALLOC(ch2d,CH2D_T,Nz+1) ||
	    MALLOC(A,char,Ny*Nx) ||
	    MALLOC(B,char,Ny*Nx) ||
	    MALLOC(C,Cell *,Ny) ||
	    MALLOC(*C,Cell,Ny*Nx) ||
	    MALLOC(Xl,int,Ny+1) ||
	    MALLOC(Yl,int,Ny+1) ||
	    MALLOC(Xr,int,Ny+1) ||
	    MALLOC(Yr,int,Ny+1)) Error(mae);

	for (y=1; y<Ny; y++) C[y]=C[y-1]+Nx;

	n=0;
	for (z=0; z<=Nz; z++) {
	    if (z==Nz)
		a=b=(z%2==0)?B:A;
	    else {
		ReadSliceImage(paths[z],Nx,Ny,C);

		if (z==0)
		    a=b=A;
		else if (z%2==0) {
		    a=A; b=B;
		}
		else {
		    a=B; b=A;
		}
		for (W=a, y=0; y<Ny; y++)
		for (x=0; x<Nx; x++, W++) *W=flag[C[y][x]];
	    }
	    i=0;
	    for (y=0; y<Ny; y++) {
		l=Nx;
		for (x=0; x<Nx; x++)
		    if (*(a++)|*(b++) && (r=x)<l) l=x;

		if (l!=Nx)
		    if (i==0) {
			Xl[0]=Xl[1]=l; Xr[0]=Xr[1]= -r;
			Yl[0]=Yr[0]=y; Yl[1]=Yr[1]=y+1; L=R=i=1;
		    }
		    else {
			L=Append(L,Xl,Yl, l,y);
			R=Append(R,Xr,Yr,-r,y);
		    }
	    }
	    if (i==0)
		ch2d[z].n=0;
	    else {
		n+=(i=ch2d[z].n=L+R+2);
		if (MALLOC(Xz=ch2d[z].x,int,i) ||
		    MALLOC(Yz=ch2d[z].y,int,i)) Error(mae);

		for (i=0; i<=R; i++) {
		    *(Xz++)=1-Xr[i]; *(Yz++)=Yr[i];
		}
		for (i=L; i>=0; i--) {
		    *(Xz++)=  Xl[i]; *(Yz++)=Yl[i];
		}
	    }
	}
	free(Yr); free(Xr);
	free(Yl); free(Xl); free(*C); free(C); free(B); free(A);

	ch3d=InitCH3D(n); X=CH3D_X(ch3d); Y=CH3D_Y(ch3d); Z=CH3D_Z(ch3d);

	n=0;
	for (z=0; z<=Nz; z++)
	    if ((i=ch2d[z].n)!=0) {
		while (--i>=0) {
		    X[n]=ch2d[z].x[i]; Y[n]=ch2d[z].y[i]; Z[n++]=z;
		}
		free(ch2d[z].y); free(ch2d[z].x);
	    }

	free(ch2d);

	CH3D(ch3d,n,0);

	if (argv[4][0]=='-')
	    file=stdout;
	else
	    if ((file=fopen(argv[4],"wb"))==NULL) Error("file not stored.");

	(void)fprintf(file,"ply\n"
			   "format binary_little_endian 1.0\n"
			   "element vertex %lu\n"
			   "property float x\n"
			   "property float y\n"
			   "property float z\n"
			   "element face %lu\n"
			   "property uchar red\n"
			   "property uchar green\n"
			   "property uchar blue\n"
			   "property list uchar int vertex_indices\n"
			   "end_header\n",
			   (unsigned long)ch3d->Nv1,
			   (unsigned long)ch3d->Nf1);

	for (n=0; n<ch3d->Nv1; n++) {
	    p=CH3D_V2P(ch3d,n); f=(float)X[p]; (*Put)(file,&f);
				f=(float)Y[p]; (*Put)(file,&f);
				f=(float)Z[p]; (*Put)(file,&f);
	}
	for (n=0; n<ch3d->Nf1; n++) {
	    CH3D_f2p(ch3d,n,P);
	    (void)fputc(Normal(P,Y,Z)*255,file);
	    (void)fputc(Normal(P,Z,X)*255,file);
	    (void)fputc(Normal(P,X,Y)*255,file);
	    (void)fputc(3,file);
	    u=(unsigned)CH3D_P2V(ch3d,P[0]); (*Put)(file,&u);
	    u=(unsigned)CH3D_P2V(ch3d,P[1]); (*Put)(file,&u);
	    u=(unsigned)CH3D_P2V(ch3d,P[2]); (*Put)(file,&u);
	}
	if (file!=stdout) (void)fclose(file);

	TermCH3D(ch3d); return 0;
}
