
#include <stdio.h>
#include <stdlib.h>
#include "cell.h"
#include "csi.h"
#include "ch3d.h"

extern void	Error(char *msg),
		RangeList(char *str,size_t lim,char *flag),
		ReadSliceImage(char *path,int Nx,int Ny,Cell **cell);

static int	Append(int i2,int *X,int *Y,int x,int y)
{
	int	i1,x1,y1,x2=X[i2]-x,y2=Y[i2]-y;

	while ((i1=i2)>0) {
	    --i2; x1=x2; y1=y2; if ((x2=X[i2]-x)*y1<(y2=Y[i2]-y)*x1) break;
	}
	++i1; X[i1]=x; Y[i1]=y++; x1=0; y1=(-1);
	while ((i2=i1)>0) {
	    --i1; x2=x1; y2=y1; if ((x1=X[i1]-x)*y2<(y1=Y[i1]-y)*x2) break;
	}
	++i2; X[i2]=x; Y[i2]=y; return i2;
}

#define PV	(Cell)(~0)

typedef struct LIST {
		int		I1,I2,*X1,*Y1,*X2,*Y2,z;
		struct LIST	*next;
	} List;

typedef struct {
		CH3D_U	n;
		List	*list,**next;
	} Table;

#define MALLOC(var,type,noe)	(var=(type *)malloc(sizeof(type)*(noe)))==NULL
#define REALLOC_INT(var,noe)	\
	(var=(int *)realloc(var,sizeof(int)*(noe)))==NULL

int	main(int argc,char **argv)
{
	char			**path,*mae="memory allocation error.";
	int			Nx,Ny,Nz,y,z,i,x;
	static char		flag[PV+1];
	Cell			**cell,**last;
	unsigned long		c,n1=0,n2=0;
	static int		X1[PV+1],X2[PV+1];
	static List		*L[PV+1];
	static Table		T[PV+1];
	static unsigned long	S1[PV+1],S2[PV+1];
	List			*l,*next;
	CH3D_U			n,p,P[3];
	CH3D_T			*ch3d;
	CH3D_I			*X,*Y,*Z;

	if (argc!=4) Error("usage : si_chs directory nameFile rangeList");

	path=CheckSliceImages(argv[1],argv[2],&Nx,&Ny,&Nz,NULL);

	RangeList(argv[3],PV,flag);

	if (MALLOC(cell,Cell *,Ny) ||
	    MALLOC(*cell,Cell,Ny*Nx) ||
	    MALLOC(last,Cell *,Ny) ||
	    MALLOC(*last,Cell,Ny*Nx)) Error(mae);

	for (y=1; y<Ny; y++) {
	    cell[y]=cell[y-1]+Nx; last[y]=last[y-1]+Nx;
	}
	for (c=0; c<=PV; c++) {
	    X1[c]=Nx; L[c]=NULL; T[c].n=0; T[c].next=(&(T[c].list));
	}
	for (z=0; z<=Nz; z++) {
	    if (z==Nz)
		i=1;
	    else {
		ReadSliceImage(path[z],Nx,Ny,cell); free(path[z]); i=z==0;
	    }
	    for (y=0; y<Ny; y++) {
		for (x=0; x<Nx; x++) {
		    if (i) {
			if (flag[c=cell[y][x]]) {
			    X2[c]=x; if (X1[c]==Nx) X1[S1[n1++]=c]=x;
			}
		    }
		    else if (last[y][x]!=cell[y][x]) {
			if (flag[c=last[y][x]]) {
			    X2[c]=x; if (X1[c]==Nx) X1[S1[n1++]=c]=x;
			}
			if (flag[c=cell[y][x]]) {
			    X2[c]=x; if (X1[c]==Nx) X1[S1[n1++]=c]=x;
			}
		    }
		    last[y][x]=cell[y][x];
		}
		while (n1>0) {
		    if ((l=L[c=S1[--n1]])==NULL) {
			if (MALLOC(l=L[c],List,1) ||
			    MALLOC(l->X1,int,Ny+1-y) ||
			    MALLOC(l->Y1,int,Ny+1-y) ||
			    MALLOC(l->X2,int,Ny+1-y) ||
			    MALLOC(l->Y2,int,Ny+1-y)) Error(mae);

			l->X1[0]=l->X1[1]=  X1[c]; l->Y1[0]=l->Y2[0]=y  ;
			l->X2[0]=l->X2[1]= -X2[c]; l->Y1[1]=l->Y2[1]=y+1;
			l->I1=l->I2=1; l->z=z; S2[n2++]=c;
		    }
		    else {
			l->I1=Append(l->I1,l->X1,l->Y1, X1[c],y);
			l->I2=Append(l->I2,l->X2,l->Y2,-X2[c],y);
		    }
		    X1[c]=Nx;
		}
	    }
	    while (n2>0) {
		l=L[c=S2[--n2]];
		if (REALLOC_INT(l->X1,l->I1+1) ||
		    REALLOC_INT(l->Y1,l->I1+1) ||
		    REALLOC_INT(l->X2,l->I2+1) ||
		    REALLOC_INT(l->Y2,l->I2+1)) Error(mae);

		T[c].n+=(l->I1+l->I2+2);
		*(T[c].next)=l; *(T[c].next=(&(l->next)))=L[c]=NULL;
	    }
	}
	free(*last); free(last); free(*cell); free(cell); free(path);

	n=0; for (c=0; c<=PV; c++) if (T[c].n>n) n=T[c].n;

	if (n!=0)
{
	ch3d=InitCH3D(n); X=CH3D_X(ch3d); Y=CH3D_Y(ch3d); Z=CH3D_Z(ch3d);

	for (c=0; c<=PV; c++) if (T[c].n!=0)
{
	n=0;
	for (l=T[c].list; l!=NULL; l=next) {
	    for (i=0; i<=l->I1; i++) {
		X[n]=  l->X1[i]; Y[n]=l->Y1[i]; Z[n++]=l->z;
	    }
	    for (i=l->I2; i>=0; i--) {
		X[n]=1-l->X2[i]; Y[n]=l->Y2[i]; Z[n++]=l->z;
	    }
	    free(l->Y2); free(l->X2);
	    free(l->Y1); free(l->X1); next=l->next; free(l);
	}
	CH3D(ch3d,n,0);

	(void)printf("%lu\t%lu\t%lu\n",c,(unsigned long)ch3d->Nv1,
					 (unsigned long)ch3d->Nf1);

	for (n=0; n<ch3d->Nv1; n++) {
	    p=CH3D_V2P(ch3d,n);
	    (void)printf("%d\t%d\t%d\n",(int)X[p],(int)Y[p],(int)Z[p]);
	}
	for (n=0; n<ch3d->Nf1; n++) {
	    CH3D_f2p(ch3d,n,P);
	    (void)printf("%lu\t%lu\t%lu\n",(unsigned long)CH3D_P2V(ch3d,P[0]),
					   (unsigned long)CH3D_P2V(ch3d,P[1]),
					   (unsigned long)CH3D_P2V(ch3d,P[2]));
	}
}
	TermCH3D(ch3d);
}
	return 0;
}
