
#include <stdio.h>
#include <stdlib.h>
#include "cell.h"
#include "csi.h"

extern void	Error(char *msg),
		RangeList(char *str,size_t lim,char *obj),
		ReadSliceImage(char *path,int Nx,int Ny,Cell **slice),
		chull3_(int *V,int *X,int *Y,int *Z,int *name,int *para,
			int *ve,int *fe,int *sv,int *ev,int *rf,int *lf,
			int *sce,int *scce,int *ece,int *ecce,
			int *wv1,int *wv2,int *wv3,int *wv4,int *wv5,int *wv6,
			int *we1,int *we2,int *wf1,int *wf2);

static int	Append(int i2,int *X,int *Y,int x,int y)
{
	int	i1,x1,y1,x2=X[i2]-x,y2=Y[i2]-y;

	while ((i1=i2)>0) {
	    --i2; x1=x2; y1=y2; if ((x2=X[i2]-x)*y1<(y2=Y[i2]-y)*x1) break;
	}
	++i1; X[i1]=x; Y[i1]=y++; x1=0; y1=(-1);
	while ((i2=i1)>0) {
	    --i1; x2=x1; y2=y1; if ((x1=X[i1]-x)*y2<(y1=Y[i1]-y)*x2) break;
	}
	++i2; X[i2]=x; Y[i2]=y; return i2;
}

static int	FV3(int *X,int *Y,int *Z,
		    int *fe,int *sv,int *ev,int *rf,int *sce,int *ece,
		    int f,int *v1,int *v2,int *v3)
{
	int		e=fe[f]-1;
	long long	x1,y1,z1,x2,y2,z2;

	if (f==rf[e]-1) {
	    *v1=ev[e]-1; *v2=sv[e]-1; e=sce[e]-1;
	}
	else {
	    *v1=sv[e]-1; *v2=ev[e]-1; e=ece[e]-1;
	}
	*v3=(f==rf[e]-1)?sv[e]-1:ev[e]-1;

	return (x1=X[*v1]-X[*v3])*(y2=Y[*v2]-Y[*v3])!=
	       (y1=Y[*v1]-Y[*v3])*(x2=X[*v2]-X[*v3]) ||
	       (z1=Z[*v1]-Z[*v3])*x2!=x1*(z2=Z[*v2]-Z[*v3]) || y1*z2!=z1*y2;
}

#define PV	(Cell)(~0)

typedef struct LIST {
		int    		I1,I2,*X1,*Y1,*X2,*Y2,z;
		struct LIST	*next;
	} List;

typedef struct {
		int	V;
		List	*list,**ptr;
	} Table;

int	main(int argc,char **argv)
{
	char		**path;
	int		Nx,Ny,Nz,y,n1,n2,i,z,x,j,V,k,*X,*Y,*Z,
			*name,*ve,*fe,*sv,*ev,*rf,*lf,*sce,*scce,*ece,*ecce,
			*wv1,*wv2,*wv3,*wv4,*wv5,*wv6,*we1,*we2,*wf1,*wf2,
			x1,x2,para[10],F,v1,v2,v3;
	static char	object[PV+1];
	Cell		**cell,**last,c;
	static int	X1[PV+1],X2[PV+1];
	static List	*L[PV+1];
	static Table	T[PV+1];
	static Cell	S1[PV+1],S2[PV+1];
	List		*l;

	if (argc!=4) Error("usage : si_chs directory nameFile rangeList");

	path=CheckSliceImages(argv[1],argv[2],&Nx,&Ny,&Nz,NULL);

	RangeList(argv[3],PV,object);

	if ((cell=(Cell **)malloc(sizeof(Cell *)*Ny))==NULL ||
	    (*cell=(Cell *)malloc(sizeof(Cell)*Ny*Nx))==NULL ||
	    (last=(Cell **)malloc(sizeof(Cell *)*Ny))==NULL ||
	    (*last=(Cell *)malloc(sizeof(Cell)*Ny*Nx))==NULL)
	  Error("no allocatable memory for image.");

	for (y=1; y<Ny; y++) {
	  cell[y]=cell[y-1]+Nx; last[y]=last[y-1]+Nx;
	}
	n1=n2=0;
	for (i=0; i<=PV; i++) {
	  X1[i]=Nx; L[i]=NULL; T[i].V=0; T[i].ptr=(&(T[i].list));
	}
	for (z=0; z<=Nz; z++) {
	  if (z<Nz) {
	    ReadSliceImage(path[z],Nx,Ny,cell); free(path[z]);
	  }
	  for (y=0; y<Ny; y++) {
	    for (x=0; x<Nx; x++) {
	      if (z>0 && object[c=last[y][x]]) {
		X2[c]=x; if (X1[c]==Nx) X1[S1[n1++]=c]=x;
	      }
	      if (z<Nz && object[c=last[y][x]=cell[y][x]]) {
		X2[c]=x; if (X1[c]==Nx) X1[S1[n1++]=c]=x;
	      }
	    }
	    while (n1>0) {
	      if ((l=L[c=S1[--n1]])==NULL) {
		if ((l=L[c]=(List *)malloc(sizeof(List)))==NULL ||
		    (l->X1=(int *)malloc(sizeof(int)*(i=Ny+1-y)))==NULL ||
		    (l->Y1=(int *)malloc(sizeof(int)*i))==NULL ||
		    (l->X2=(int *)malloc(sizeof(int)*i))==NULL ||
		    (l->Y2=(int *)malloc(sizeof(int)*i))==NULL)
		  Error("no allocatable memory for 2D convex hull.");

		l->I1=l->I2=1;
		l->X1[0]=l->X1[1]=  X1[c]; l->Y1[0]=l->Y2[0]=y;
		l->X2[0]=l->X2[1]= -X2[c]; l->Y1[1]=l->Y2[1]=y+1;

		S2[n2++]=c;
	      }
	      else {
		l->I1=Append(l->I1,l->X1,l->Y1, X1[c],y);
		l->I2=Append(l->I2,l->X2,l->Y2,-X2[c],y);
	      }
	      X1[c]=Nx;
	    }
	  }
	  while (n2>0) {
	    l=L[c=S2[--n2]];
	    if ((l->X1=(int *)realloc(l->X1,sizeof(int)*(i=l->I1+1)))==NULL ||
		(l->Y1=(int *)realloc(l->Y1,sizeof(int)*i))==NULL ||
		(l->X2=(int *)realloc(l->X2,sizeof(int)*(j=l->I2+1)))==NULL ||
		(l->Y2=(int *)realloc(l->Y2,sizeof(int)*j))==NULL)
	      Error("memory re-allocation error.");

	    l->z=z; l->next=L[c]=NULL;

	    T[c].V+=(i+j); *(T[c].ptr)=l; T[c].ptr=(&(l->next));
	  }
	}
	free(path); free(*last); free(last); free(*cell); free(cell);

	V=0; for (i=0; i<=PV; i++) if (T[i].V>V) V=T[i].V;

	if (V!=0) {
	  i=V*2+1; j=i*3; k=i*4;
	  if ((X=(int *)malloc(sizeof(int)*i))==NULL ||
	      (Y=(int *)malloc(sizeof(int)*i))==NULL ||
	      (Z=(int *)malloc(sizeof(int)*i))==NULL ||
	      (name=(int *)malloc(sizeof(int)*i))==NULL ||
	      (ve=(int *)malloc(sizeof(int)*i))==NULL ||
	      (fe=(int *)malloc(sizeof(int)*j))==NULL ||
	      (sv=(int *)malloc(sizeof(int)*k))==NULL ||
	      (ev=(int *)malloc(sizeof(int)*k))==NULL ||
	      (rf=(int *)malloc(sizeof(int)*k))==NULL ||
	      (lf=(int *)malloc(sizeof(int)*k))==NULL ||
	      (sce=(int *)malloc(sizeof(int)*k))==NULL ||
	      (scce=(int *)malloc(sizeof(int)*k))==NULL ||
	      (ece=(int *)malloc(sizeof(int)*k))==NULL ||
	      (ecce=(int *)malloc(sizeof(int)*k))==NULL ||
	      (wv1=(int *)malloc(sizeof(int)*i))==NULL ||
	      (wv2=(int *)malloc(sizeof(int)*i))==NULL ||
	      (wv3=(int *)malloc(sizeof(int)*i))==NULL ||
	      (wv4=(int *)malloc(sizeof(int)*i))==NULL ||
	      (wv5=(int *)malloc(sizeof(int)*i))==NULL ||
	      (wv6=(int *)malloc(sizeof(int)*i))==NULL ||
	      (we1=(int *)malloc(sizeof(int)*k))==NULL ||
	      (we2=(int *)malloc(sizeof(int)*k))==NULL ||
	      (wf1=(int *)malloc(sizeof(int)*j))==NULL ||
	      (wf2=(int *)malloc(sizeof(int)*j))==NULL)
	    Error("no allocatable memory for 3D convex hull.");

	  for (i=0; i<=PV; i++)
	    if (T[i].V!=0) {
	      V=0;
	      for (l=T[i].list; l!=NULL; l=l->next) {
		z=l->z;
		for (x1=x2=0; x1<=l->I1 || x2<=l->I2; V++) {
		  if (x2>l->I2 || (x1<=l->I1 && l->Y1[x1]<=l->Y2[x2])) {
		    X[V]=  l->X1[x1]; Y[V]=l->Y1[x1]; ++x1;
		  }
		  else {
		    X[V]=1-l->X2[x2]; Y[V]=l->Y2[x2]; ++x2;
		  }
		  Z[name[V]=V]=z;
		}
	      }
	      para[0]=n1=V+1; chull3_(&V,X,Y,Z,name,para,
				      ve,fe,sv,ev,rf,lf,sce,scce,ece,ecce,
				      wv1,wv2,wv3,wv4,wv5,wv6,we1,we2,wf1,wf2);
	      n2=n1*3;

	      V=0;
	      for (j=0; j<n1; j++) {
		name[j]=V; V+=wv5[j];
	      }
	      F=0;
	      for (j=0; j<n2; j++)
		F+=(wf1[j] && FV3(X,Y,Z,fe,sv,ev,rf,sce,ece,j,&v1,&v2,&v3));

	      (void)printf("%d\t%d\t%d\n",i,V,F);

	      for (j=0; j<n1; j++)
		if (wv5[j]) (void)printf("%d\t%d\t%d\n",X[j],Y[j],Z[j]);

	      for (j=0; j<n2; j++)
		if (wf1[j] && FV3(X,Y,Z,fe,sv,ev,rf,sce,ece,j,&v1,&v2,&v3))
		  (void)printf("%d\t%d\t%d\n",name[v1],name[v2],name[v3]);
	    }
	}
	return 0;
}
