
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include "cell.h"
#include "rif.h"

extern void	Error(char *msg),
		RangeList(char *str,size_t lim,char *obj);

static int	Append(int i2,int *X,int *Y,int x,int y)
{
	int	i1,x1,y1,x2=X[i2]-x,y2=Y[i2]-y;

	while ((i1=i2)>0) {
	    --i2; x1=x2; y1=y2; if ((x2=X[i2]-x)*y1<(y2=Y[i2]-y)*x1) break;
	}
	++i1; X[i1]=x; Y[i1]=y++; x1=0; y1=(-1);
	while ((i2=i1)>0) {
	    --i1; x2=x1; y2=y1; if ((x1=X[i1]-x)*y2<(y1=Y[i1]-y)*x2) break;
	}
	++i2; X[i2]=x; Y[i2]=y; return i2;
}

static double	Square(int i)
{
	return (double)i*(double)i;
}

#define PV	(size_t)(Cell)(~0)

typedef struct {
		int     H,L,R,*X,*Y;
	} Table;

int	main(int argc,char **argv)
{
	int		Nx,Ny,n,i,y,x,j,
			k,o,x1,y1,tx,x3,x2,ty,y3,y2,x0,dx,y0,dy,a1,a2,b1,b2;
	Cell		**cell,c;
	static char	object[PV+1];
	static int	L[PV+1],R[PV+1];
	static Table	*T[PV+1];
	static Cell	S[PV+1];
	Table		*t;
	double		B,W,A,Bx,By,Ax,Ay,B0,A0,B1,B2,A1,A2,A3,B3;

	if (argc!=3) Error("usage : t_bbs TIFF rangeList");

	ReadImageFile(argv[1],&Nx,&Ny,NULL,&cell,NULL);

	RangeList(argv[2],PV,object);

	n=0;
	for (i=0; i<=PV; i++) {
	    L[i]=Nx; T[i]=NULL;
	}
	for (y=0; y<Ny; y++) {
	    for (x=0; x<Nx; x++)
		if (object[c=cell[y][x]]) {
		    R[c]=x; if (L[c]==Nx) L[S[n++]=c]=x;
		}

	    while (n>0) {
		if ((t=T[c=S[--n]])==NULL) {
		    if ((t=T[c]=(Table *)malloc(sizeof(Table)))==NULL ||
			(t->X=(int *)malloc(sizeof(int)*(j=(i=Ny+1-y)*2
							)))==NULL ||
			(t->Y=(int *)malloc(sizeof(int)*j))==NULL)
			Error("memory allocation error.");

		    t->H=i;
		    t->L=t->R=1;
		    t->X[0]=t->X[    1]=  L[c]; t->Y[0]=t->Y[i]=y  ;
		    t->X[i]=t->X[j=i+1]= -R[c]; t->Y[1]=t->Y[j]=y+1;
		}
		else {
		    t->L=Append(t->L,t->X     ,t->Y	, L[c],y);
		    t->R=Append(t->R,t->X+t->H,t->Y+t->H,-R[c],y);
		}
		L[c]=Nx;
	    }
	}
	for (i=0; i<=PV; i++)
	    if ((t=T[i])!=NULL) {
		for (k=(j=t->H)+t->R; j<k; j++, k--) {
		    x=t->X[j]; t->X[j]=t->X[k]; t->X[k]=x;
		    y=t->Y[j]; t->Y[j]=t->Y[k]; t->Y[k]=y;
		}
		for (n=t->L+1, j=t->H, k=0; k<=t->R; k++, j++, n++) {
		    t->X[n]=1-t->X[j]; t->Y[n]=t->Y[j];
		}
		o=n-1; x1=t->X[o];
		       y1=t->Y[o]; tx=(x3=t->X[1])-(x2=t->X[0]);
				   ty=(y3=t->Y[1])-(y2=t->Y[0]); j=1; B=0.0;
		for (k=0; k<n; k++) {
		    x0=x1; dx=(x1=t->X[k])-x0;
		    y0=y1; dy=(y1=t->Y[k])-y0;
		    while (dy*tx-dx*ty>0) {
			j=(j+1)%n; x2=x3; tx=(x3=t->X[j])-x2;
				   y2=y3; ty=(y3=t->Y[j])-y2;
		    }

#define S(x,y)	(Square(x)+Square(y))

		    W=Square(dy*(x0-=x2)-dx*(y0-=y2))/S(dx,dy);
		    if (k==0 || W<A) {
			A=W; a1=k; a2=j;
		    }
		    if ((W=S(x0,y0))>B) {
			B=W; b1=k; b2=j;
		    }
		    if ((W=S(x1-x2,y1-y2))>B) {
			B=W; b1=k+1; b2=j;
		    }
		}
		x0=t->X[a1]; y0=t->Y[a1]; a1=(a1+o)%n;
		W=sqrt(S(tx=x0-t->X[a1],ty=y0-t->Y[a1])); Bx=(double)tx/W;
							  By=(double)ty/W;

		a2=(a2+o)%n; x1=t->X[a2]; y1=t->Y[a2];

		b1=(b1+o)%n; x2=t->X[b1]; y2=t->Y[b1];
		b2=(b2+o)%n; x3=t->X[b2]; y3=t->Y[b2];
		W=sqrt(S(dx=x3-x2,dy=y3-y2)); Ax=  (double)dy/W;
					      Ay= -(double)dx/W;

		for (j=0; j<n; j++) {
		    x=t->X[j];
		    y=t->Y[j]; B0=Bx*(double)x+By*(double)y;
			       A0=Ax*(double)x+Ay*(double)y;
		    if (j==0) {
			B1=B2=B0; A1=A2=A0;
		    }
		    else {
			if (B0<B1) B1=B0; else if (B0>B2) B2=B0;
			if (A0<A1) A1=A0; else if (A0>A2) A2=A0;
		    }
		}
		A0=  By*(double)x0-Bx*(double)y0;
		A3=  By*(double)x1-Bx*(double)y1;
		B0= -Ay*(double)x2+Ax*(double)y2;
		B3= -Ay*(double)x3+Ax*(double)y3;
		(void)printf("%d\n"
			     "%lg\t%lg\n"
			     "%lg\t%lg\t%lg\t%lg\t%lg\t%lg\n"
			     "%lg\t%lg\n"
			     "%lg\t%lg\t%lg\t%lg\t%lg\t%lg\n",
			     i,sqrt(A),B2-B1,By*A0+Bx*B1,-Bx*A0+By*B1,
					     By*A3+Bx*B1,-Bx*A3+By*B1,
					     By*A0+Bx*B2,-Bx*A0+By*B2,
			       A2-A1,sqrt(B),Ax*A1-Ay*B0, Ay*A1+Ax*B0,
					     Ax*A2-Ay*B0, Ay*A2+Ax*B0,
					     Ax*A1-Ay*B3, Ay*A1+Ax*B3);
	    }

	return 0;
}
