
#include <stdio.h>
#include <stdlib.h>
#include "cell.h"
#include "rif.h"
#include "sif_cm.h"
#include "sif.h"

#define EPS	1e-9	/* ==DBL_EPSILON */

#define LABEL	unsigned long
#define FORMAT	"%lu"

static unsigned short	cmR[4]={65535,65535,0,0},
			cmG[4]={65535,0,65535,0},
			cmB[4]={65535,0,0,65535};

extern void	Error(),RangeList();

static LABEL	Trace(link,l1)
LABEL		*link,l1;
{
	LABEL	l2=link[l1];

	return (l2==0)?l1:(link[l1]=Trace(link,l2));
}

static int	Append(i2,X,Y,x,y)
int		i2,*X,*Y,x,y;
{
	int	i1,x1,y1,x2=X[i2]-x,y2=Y[i2]-y;

	while ((i1=i2)>0) {
	    --i2; x1=x2; y1=y2; if ((x2=X[i2]-x)*y1<(y2=Y[i2]-y)*x1) break;
	}
	++i1; X[i1]=x; Y[i1]=y++; x1=0; y1=(-1);
	while ((i2=i1)>0) {
	    --i1; x2=x1; y2=y1; if ((x1=X[i1]-x)*y2<(y1=Y[i1]-y)*x2) break;
	}
	++i2; X[i2]=x; Y[i2]=y; return i2;
}

#define PV	(size_t)(Cell)(~0)

int	main(argc,argv)
int	argc;
char	**argv;
{
	int	Nx,Ny,I,y,*X1,*Y1,*X2,*Y2,x,y1,x1,x2,I1,I2,y2,x1i,y1i,x2i,y2i;
	Cell	**cell;
	char	pv[PV+1];
	LABEL	*last,*link,L,l,lx,CL=0,C3=0,C2=0,C1=0;
	double	x1d,y1d,g1,x2d,y2d,g2;

	if (argc!=3 && argc!=4)
	    Error("usage : t_ch orgTIFF rangeList {newTIFF}");

	ReadImageFile(argv[1],&Nx,&Ny,NULL,&cell,NULL);

	RangeList(argv[2],PV,pv);

	I=Nx+1;

	y=Ny+1;
	if ((X1=(int *)malloc(sizeof(int)*y))==NULL ||
	    (Y1=(int *)malloc(sizeof(int)*y))==NULL ||
	    (X2=(int *)malloc(sizeof(int)*y))==NULL ||
	    (Y2=(int *)malloc(sizeof(int)*y))==NULL ||
	    (last=(LABEL *)malloc(sizeof(LABEL)*I))==NULL ||
	    (link=(LABEL *)malloc(sizeof(LABEL)*I))==NULL)
	    Error("memory allocation error.");

	++last; last[-1]=1; link[1]=0;

	L=2; for (x=0; x<Nx; x++) last[x]=1;

	y1=Ny;
	for (y=0; y<Ny; y++) {
	    x1=Nx;
	    for (x=0; x<Nx; x++)
		if (pv[cell[y][x]]) {
		    if ((x2=x)<x1) x1=x;

		    last[x]=0;
		}
		else if ((l=last[x])!=0) {
		    l=Trace(link,l);
		    if ((lx=last[x-1])!=0 && (lx=Trace(link,lx))!=l)
			if (lx==1) l=link[l]=1; else link[lx]=l;

		    last[x]=l;
		}
		else if ((l=last[x-1])!=0)
		    last[x]=Trace(link,l);
		else {
		    if (L%I==0 &&
		       (link=(LABEL *)realloc(link,sizeof(LABEL)*(L+I)))==NULL)
			Error("memory re-allocation error.");

		    link[last[x]=L++]=0;
		}

	    if (x1!=Nx) {
		y2=y+1;
		if (y1==Ny) {
		    X1[0]=X1[1]=  x1; Y1[0]=Y2[0]=y1=y;
		    X2[0]=X2[1]= -x2; Y1[1]=Y2[1]=y2  ; I1=I2=1;
		}
		else {
		    I1=Append(I1,X1,Y1, x1,y);
		    I2=Append(I2,X2,Y2,-x2,y);

		    if (Nx>1 &&
		       (l=last[Nx-1])!=0 && (l=Trace(link,l))!=1) link[l]=1;
		}
	    }
	}
	if (y1!=Ny) {
	    if (Ny>1)
		for (x=Nx-1; x>0; x--)
		    if ((l=last[x])!=0 && (l=Trace(link,l))!=1) link[l]=1;
	
	    for (l=2; l<L; l++) if (link[l]!=0) (void)Trace(link,l);
	    for (l=2; l<L; l++)
		if (link[l]==0) {
		    link[l]=l; ++CL;
		}

	    L=2; for (x=0; x<Nx; x++) last[x]=1;

	    x1i=X1[0]; y1i=Y1[0];
	    x2i=X2[0]; y2i=Y2[0]; I1=I2=0;
	    for (y=y1; y<y2; y++) {
		if (y>=y1i) {
		    x1d=(double)x1i;
		    y1d=(double)y1i; ++I1; g1=((double)(x1i=X1[I1])-x1d)/
					      ((double)(y1i=Y1[I1])-y1d);
		}
		if (y>=y2i) {
		    x2d=(double)x2i;
		    y2d=(double)y2i; ++I2; g2=(x2d-(double)(x2i=X2[I2]))/
					      ((double)(y2i=Y2[I2])-y2d);
		}
		x1=(int)(g1*((double)y-y1d+0.5)+x1d+0.5-EPS);
		x2=(int)(g2*((double)y-y2d+0.5)-x2d+1.5+EPS);

		for (x=0; x<x1; x++) {
		    last[x]=1; cell[y][x]=0;
		}
		for (; x<x2; x++)
		    if (pv[cell[y][x]]) {
			last[x]=0; cell[y][x]=3; ++C3;
		    }
		    else {
			if ((l=last[x])==0 && (l=last[x-1])==0) l=link[L++];

			if ((last[x]=l)!=1) {
			    cell[y][x]=2; ++C2;
			}
			else {
			    cell[y][x]=1; ++C1;
			}
		    }

		for (; x<Nx; x++) {
		    last[x]=1; cell[y][x]=0;
		}
	    }
	    if (argc==4) {
		for (y= 0; y<y1; y++) for (x=0; x<Nx; x++) cell[y][x]=0;
		for (y=y2; y<Ny; y++) for (x=0; x<Nx; x++) cell[y][x]=0;

		StoreImageFile_CM(argv[3],Nx,Ny,2,cmR,cmG,cmB,cell,NULL);
	    }
	}
	else if (argc==4) {
	    for (y=0; y<Ny; y++) for (x=0; x<Nx; x++) cell[y][x]=0;

	    StoreImageFile(argv[3],Nx,Ny,1,cell,NULL);
	}
	(void)printf(FORMAT "\n"
		     FORMAT "\t" FORMAT "\n"
		     FORMAT "\n",C1,C2,CL,C3);

	return 0;
}
